"""
Provides functions to gather usage statistics for Portal/ArcGIS Online
"""
import os
import time
import datetime
from typing import Optional
from .._impl._con import Connection
from ..._impl.common._mixins import PropertyMap
from ..._impl.common._utils import local_time_to_online, timestamp_to_datetime
from ...gis import GIS
from ._base import BasePortalAdmin


########################################################################


class AGOLUsageReports(BasePortalAdmin):
    """
    Simple Usage Reports from ArcGIS Online

    .. note::
        Usage reports can contain users outside your organization.

    """

    _json_dict = {}
    _json = None
    _con = None
    _portal = None
    _gis = None
    _url = None

    # ----------------------------------------------------------------------

    def _init(self, connection=None):
        """loads the properties into the class"""
        self._json_dict = {}
        self._properties = PropertyMap(self._json_dict)

    # ----------------------------------------------------------------------
    def generate_report(
        self,
        focus: str = "org",
        report_type: str = "users",
        title: Optional[str] = None,
        duration: Optional[str] = None,
        start_time: Optional[datetime.datetime] = None,
        notify: bool = False,
        future: bool = True,
    ):
        """
        Generates the reports of the overall usage of the organizations.
        Reports define organization usage metrics for either a weekly or
        monthly time frame.


        ===============     ====================================================
        **Parameter**        **Description**
        ---------------     ----------------------------------------------------
        focus               Optional String. The report type. Currently, only
                            the organization (`org`) report type is supported.
        ---------------     ----------------------------------------------------
        report_type         Required String. The type of report to generate.

                            Values:
                                - 'content'
                                - 'users'
                                - 'activity'
                                - 'credits'
                                - 'serviceUsages'
                                - 'itemUsages'
        ---------------     ----------------------------------------------------
        title               *deprecated* Optional String.  The output report item's title.
        ---------------     ----------------------------------------------------
        duration            Optional String. Specifies the time duration for the
                            reports. This parameter is required when `report_type`
                            is set to `credits`, `activity`, `serviceUsages`, or `itemUsages`.

                            .. note::
                                The `daily` value is only available when `report_type` is
                                set to `activity`.
                                The `yearly` value is only available when `report_type`
                                is set to `itemUsages`.

                            Values:
                                - 'daily'
                                - 'weekly'
                                - 'monthly'
                                - 'quarterly'
                                - 'yearly'
        ---------------     ----------------------------------------------------
        start_time          Optional datetime.datetime. The start time of the
                            time duration. The time format is Unix time with millisecond
                            precision. If `duration = 'weekly'`, the start_time
                            value must be a time on Sunday or Monday GMT.
                            If `duration = 'monthly`, the start_time value must
                            be on the first day of the month.
        ---------------     ----------------------------------------------------
        notify              Optional Boolean. The Job will print a message upon
                            task completion.
        ---------------     ----------------------------------------------------
        future              Optional Boolean. Returns an asynchronous Job when
                            `True`, when `False`, returns an :class:`~arcgis.gis.Item`.
        ===============     ====================================================


        :return: Async Job Object or :class:`~arcgis.gis.Item`

        """
        url = f"{self._gis._portal.resturl}community/users/{self._gis.users.me.username}/report"
        params = {"f": "json", "reportType": focus, "reportSubType": report_type}

        # Perform Checks
        if duration and duration.lower() not in [
            "daily",
            "weekly",
            "monthly",
            "quarterly",
            "yearly",
        ]:
            raise ValueError("Invalid `duration` value %s" % duration)
        if duration:
            duration = duration.lower()
            if duration == "daily" and report_type != "activity":
                raise ValueError(
                    "Duration set to 'daily' can only be used with report type 'activity'."
                )
            elif duration == "yearly" and report_type != "itemUsages":
                raise ValueError(
                    "Duration set to 'yearly' can only be used with report type 'itemUsages'."
                )
            params["timeDuration"] = duration
        if (
            report_type in ["credits", "activity", "serviceUsages", "itemUsages"]
            and duration is None
        ):
            raise ValueError(
                "For the report type specified, a duration must also be specified."
            )

        # Assign parameters
        if not start_time is None and isinstance(start_time, datetime.datetime):
            params["startTime"] = local_time_to_online(start_time)
        elif not start_time is None and isinstance(start_time, int):
            params["startTime"] = start_time
        count = 0
        resp = self._con.post(url, params)
        if "itemId" in resp and future:
            from arcgis._impl._async.jobs import ItemStatusJob

            item = self._gis.content.get(resp["itemId"])
            while item is None:
                time.sleep(1)
                item = self._gis.content.get(resp["itemId"])
                if count == 15:
                    raise Exception(
                        "The report cannot be generated, please resubmit the operation."
                    )
                count += 1
            item = self._gis.content.get(resp["itemId"])
            isj = ItemStatusJob(
                item=item, task_name="Generate Report", notify=notify, gis=self._gis
            )
            if future:
                return isj
            return isj.result()
        return resp

    # ----------------------------------------------------------------------
    def credit(
        self,
        start_time: Optional[datetime.datetime] = None,
        time_frame: str = "week",
        export: bool = False,
    ):
        """
        Creates a Panda's dataframe or CSV file reporting on credit consumption
        within an ArcGIS Online organization.

        ===============     ====================================================
        **Parameter**        **Description**
        ---------------     ----------------------------------------------------
        start_time          optional datetime, the time to step back from.  If
                            None, the current time is used.
        ---------------     ----------------------------------------------------
        time_frame          optional string, is the timeframe report to create.
                            Allowed values: today, week (default), 14days, 30days,
                            60days, 90days, 6months, year
        ---------------     ----------------------------------------------------
        export              optional boolean, if `True`, a csv is generated from
                            the request. If `False`, a Panda's dataframe is
                            returned. Default is `False`
        ===============     ====================================================

        :return:
             string path to csv file or Panda's Dataframe (default) that
             records the total number of credits consumed per:

             * `hour` if ``time_frame`` is `today`
             * `day` if ``time_frame`` is `week`, `7days`, `14days`, `30days`,
               `60days` or `90days`
             * `week` if ``time_frame`` is `6months`
             * `month` if ``time_frame`` is `year`

        .. code-block:: python

            # Usage example

            >>> usage_reporter = gis.admin.usage_reports

            >>> usage_reporter.credit(start_time= jan2_23,
                                      time_frame= "week")

                date	credits
                ________________________________________
                0	2022-12-26 16:00:00	173.1696
                ...
                6	2023-01-01 16:00:00	177.6483

        """
        out_folder = None
        if start_time is None:
            start_time = datetime.datetime.now()
        if isinstance(start_time, datetime.datetime) == False:
            raise ValueError("start_time and end_time must be datetime objects")
        if time_frame.lower() == "today":
            end_time = start_time - datetime.timedelta(days=1)
            period = "1h"
        elif time_frame.lower() in ["7days", "week"]:
            end_time = start_time - datetime.timedelta(days=7)
            period = "1d"
        elif time_frame.lower() == "14days":
            end_time = start_time - datetime.timedelta(days=14)
            period = "1d"
        elif time_frame.lower() in ["month", "30days"]:
            end_time = start_time - datetime.timedelta(days=30)
            period = "1d"
        elif time_frame.lower() == "60days":
            end_time = start_time - datetime.timedelta(days=60)
            period = "1d"
        elif time_frame.lower() == "90days":
            end_time = start_time - datetime.timedelta(days=90)
            period = "1d"
        elif time_frame.lower() == "6months":
            end_time = start_time - datetime.timedelta(days=180)
            period = "1w"
        elif time_frame.lower() == "year":
            end_time = start_time - datetime.timedelta(days=365)
            period = "1m"
        # Convert to timestamps
        end_time = str(int(local_time_to_online(dt=end_time)))
        start_time = str(int(local_time_to_online(dt=start_time)))
        f = "json"
        if export:
            f = "csv"
        if export and (out_folder is None or os.path.isdir(out_folder) == False):
            import tempfile

            out_folder = tempfile.gettempdir()
        params = {
            "f": f,
            "startTime": end_time,
            "endTime": start_time,
            "vars": "credits,num",
            "groupby": "stype,etype",
            "period": period,
        }
        res = self._con.post(path=self._url, postdata=params)
        if export:
            return res
        elif isinstance(res, (dict, PropertyMap)):
            import pandas as pd

            data = res["data"][0]["credits"]
            for row in data:
                if isinstance(row[0], str):
                    row[0] = int(row[0])
                row[0] = timestamp_to_datetime(timestamp=row[0])

            df = pd.DataFrame.from_records(
                data=data, columns=["date", "credits"], coerce_float=True
            )
            df["credits"] = df["credits"].astype(float)
            return df
        return res

    # ----------------------------------------------------------------------
    def users(
        self, start_time: Optional[datetime.datetime] = None, time_frame: str = "week"
    ):
        """
        Creates a credit usage report for resources of an ArcGIS Online
        organization with results aggregated by specific `username` and user's
        organization id.

        .. note::
            Reports can contain users outside your organization.

        ===============     ====================================================
        **Parameter**        **Description**
        ---------------     ----------------------------------------------------
        start_time          optional datetime, the time to step back from.  If
                            None, the current time is used.
        ---------------     ----------------------------------------------------
        time_frame          optional string, is the timeframe report to create.
                            Allowed values: today, week, 14days, 30days, 60days,
                            90days, 6months, year
        ===============     ====================================================

        :return:
             dictionary reporting the number of credits consumed by users
             through this organization.

             Results are aggregated by:
               * `hour` if ``time_frame`` is `today`
               * `day` if ``time_frame`` is `week`, `7days`, `14days`, `30days`,
                 `60days` or `90days`
               * `week` if ``time_frame`` is `6months`
               * `month` if ``time_frame`` is `year`

        .. code-block:: python

            # Usage Example:

            >>> from arcgis.gis import GIS
            >>> import datetime as dt

            >>> gis = GIS(profile="my_organizational_profile")
            >>> usage_reporter = gis.admin.usage_reports

            >>> jan2_23 = dt.datetime(2023, 1, 2)
            >>> user_usg = usage_reporter.users(start_time = jan2_23,
                                                time_frame = "week")

            >>> list(user_usg.keys())
            ['startTime', 'endTime', 'period', 'data']

            >>> type(user_usg["data"])
            list

            ### The data key's value will be a list of
            ### dictionaries. Each dictionary will have varying keys.
            ### If the dictonary has no userOrgId key, that indicates
            ### a public user account.

            >>> user_usg['data'][1]
            {'username': '<user_name1>',
             'credits': [['1672099200000', '0.0'],
                         ['1672185600000', '0.0'],
                         ...
                         ['1672617600000', '2.0E-4']]}

           >>> user_usg['data'][2]
           {'username': '<user_name2>',
            'userOrgId': 'JXrNeAy8ce1q2b4l'
            'credits': [['1672099200000', '0.0'],
                        ['1672185600000', '0.0'],
                       ...
                        ['1672617600000', '0.0']]}

        """
        out_folder = None
        if start_time is None:
            start_time = datetime.datetime.now()
        if isinstance(start_time, datetime.datetime) == False:
            raise ValueError("start_time and end_time must be datetime objects")
        end_time = start_time - datetime.timedelta(days=1)
        period = "1h"
        if time_frame.lower() in ["7days", "week"]:
            end_time = start_time - datetime.timedelta(days=7)
            period = "1d"
        elif time_frame.lower() == "14days":
            end_time = start_time - datetime.timedelta(days=14)
            period = "1d"
        elif time_frame.lower() == "30days":
            end_time = start_time - datetime.timedelta(days=30)
            period = "1d"
        elif time_frame.lower() == "60days":
            end_time = start_time - datetime.timedelta(days=60)
            period = "1d"
        elif time_frame.lower() == "90days":
            end_time = start_time - datetime.timedelta(days=90)
            period = "1d"
        elif time_frame.lower() == "6months":
            end_time = start_time - datetime.timedelta(days=180)
            period = "1w"
        elif time_frame.lower() == "year":
            end_time = start_time - datetime.timedelta(days=365)
            period = "1m"
        # Convert to timestamps
        end_time = str(int(local_time_to_online(dt=end_time)))
        start_time = str(int(local_time_to_online(dt=start_time)))
        params = {
            "f": "json",
            "startTime": end_time,
            "endTime": start_time,
            "vars": "credits",
            "groupby": "username,userorgid",
            "period": period,
        }
        res = self._con.post(path=self._url, postdata=params)
        return res

    # ----------------------------------------------------------------------
    def applications(
        self, start_time: Optional[datetime.datetime] = None, time_frame: str = "week"
    ):
        """
        Creates a usage report for all registered application logins for a
        given ArcGIS Online organization.

        .. note::
            Output can contain users outside your organization
            that used organization applications

        ===============     ====================================================
        **Parameter**        **Description**
        ---------------     ----------------------------------------------------
        start_time          optional datetime, the time to step back from.  If
                            None, the current time is used.
        ---------------     ----------------------------------------------------
        time_frame          optional string, is the timeframe report to create.
                            Allowed values: today, week, 14days, 30days, 60days,
                            90days, 6months, year
        ===============     ====================================================

        :return:
             dictionary with the number of application logins grouped by
             application and username.

             Results aggregated by:

             - `hour` if ``time_frame`` is `today`
             - `day` if ``time_frame`` is `week`, `7days`, `14days`, `30days`,
               `60days` or `90days`
             - `week` if ``time_frame`` is `6months`
             - `month` if ``time_frame`` is `year`

        .. code-block:: python

            # Usage example:

            >>> import datetime as dt
            >>> from arcgis.gis import GIS

            >>> gis = GIS(profile="my_organizational_profile)
            >>> jan2_23 = dt.datetime(2023, 1, 2)

            >>> usage_reporter = gis.admin.usage_reports

            >>> usage_reporter.applications(start_time= jan2_23,
                                            time_frame="week")

            {'startTime': 1672099200000,
            'endTime': 1672704000000,
            'period': '1d',
            'data': [{'etype': 'svcusg',
                    'stype': 'applogin',
                    'username': <username 1>,
                    'userOrgId': 'JXwx ... Ok2o',
                    'appId': 'arcgisnotebooks',
                    'appOrgId': 'Ab3e ... q0o7i',
                    'num': [['1672099200000', '0'],
                            ...
                            ['1672444800000', '4'],
                            ['1672531200000', '3'],
                            ['1672617600000', '0']]},
             ...
             ...
                    {'etype': 'svcusg',
                     'stype': 'applogin',
                     'username': 'external username2',
                     'userOrgId': 'JLxMbZo4ex3kOa2o',
                     'appId': 'arcgisonline',
                     'appOrgId': 'Ab3e ... q0o7i',
                     'num': [['1672099200000', '0'],
                             ...
                             ['1672444800000', '62'],
                             ['1672531200000', '10'],
                             ['1672617600000', '0']]}]}

        """
        out_folder = None
        if start_time is None:
            start_time = datetime.datetime.now()
        if isinstance(start_time, datetime.datetime) == False:
            raise ValueError("start_time and end_time must be datetime objects")
        end_time = start_time - datetime.timedelta(days=1)
        period = "1h"
        if time_frame.lower() in ["7days", "week"]:
            end_time = start_time - datetime.timedelta(days=7)
            period = "1d"
        elif time_frame.lower() == "14days":
            end_time = start_time - datetime.timedelta(days=14)
            period = "1d"
        elif time_frame.lower() == "30days":
            end_time = start_time - datetime.timedelta(days=30)
            period = "1d"
        elif time_frame.lower() == "60days":
            end_time = start_time - datetime.timedelta(days=60)
            period = "1d"
        elif time_frame.lower() == "90days":
            end_time = start_time - datetime.timedelta(days=90)
            period = "1d"
        elif time_frame.lower() == "6months":
            end_time = start_time - datetime.timedelta(days=180)
            period = "1w"
        elif time_frame.lower() == "year":
            end_time = start_time - datetime.timedelta(days=365)
            period = "1m"
        # Convert to timestamps
        end_time = str(int(local_time_to_online(dt=end_time)))
        start_time = str(int(local_time_to_online(dt=start_time)))
        params = {
            "f": "json",
            "startTime": end_time,
            "endTime": start_time,
            "vars": "num",
            "groupby": "appId",
            "eType": "svcusg",
            "sType": "applogin",
            "period": period,
        }
        res = self._con.post(path=self._url, postdata=params)
        return res

    # ----------------------------------------------------------------------
    def _custom(
        self,
        start_time,
        end_time,
        vars=None,
        period=None,
        groupby=None,
        name=None,
        stype=None,
        etype=None,
        appId=None,
        device_id=None,
        username=None,
        app_org_id=None,
        user_org_id=None,
        host_org_id=None,
    ):
        """
        returns the usage statistics value
        """
        if (
            isinstance(start_time, datetime.datetime) == False
            or isinstance(end_time, datetime.datetime) == False
        ):
            raise ValueError("start_time and end_time must be datetime objects")

        url = self._url

        start_time = str(int(local_time_to_online(dt=start_time)))
        end_time = str(int(local_time_to_online(dt=end_time)))

        params = {
            "f": "json",
            "startTime": end_time,
            "endTime": start_time,
            "vars": vars,
            "period": period,
            "groupby": groupby,
            "name": name,
            "stype": stype,
            "etype": etype,
            "appId": appId,
            "deviceId": device_id,
            "username": username,
            "appOrgId": app_org_id,
            "userOrgId": user_org_id,
            "hostOrgId": host_org_id,
        }

        params = {key: item for key, item in params.items() if item is not None}
        return self._con.post(path=url, postdata=params)



from __future__ import annotations
from arcgis.gis._impl._con import Connection
from arcgis._impl.common._mixins import PropertyMap
from arcgis.gis import Item
from arcgis.gis.server._service import Service
from arcgis._impl.backport import cached_property
from functools import lru_cache
from typing import Optional, Union


###########################################################################
class MissionJob(object):
    """
    Represents a Single `Job` operation for Mission Server
    """

    _properties = None
    _url = None
    _con = None
    _gis = None

    # ---------------------------------------------------------------------
    def __init__(self, url: str, gis: "GIS", **kwargs) -> "MissionJob":
        self._url = url
        self._gis = gis
        self._con = gis._con

    # ---------------------------------------------------------------------
    def __str__(self):
        return f"< {self.__class__.__name__} @ {self._url} >"

    # ---------------------------------------------------------------------
    def __repr__(self):
        return f"< {self.__class__.__name__} @ {self._url} >"

    # ---------------------------------------------------------------------
    @cached_property
    def properties(self):
        """returns the properties of the resource"""

        if self._properties is None:
            try:
                self._properties = PropertyMap(self._con.get(self._url, {"f": "json"}))
            except:
                self._properties = PropertyMap(self._con.post(self._url, {"f": "json"}))
        return self._properties

    # ---------------------------------------------------------------------
    @property
    def status(self) -> object:
        """
        Returns the status

        :return: string
        """
        resp = self._con.get(self._url, {"f": "json"})
        try:
            return resp["status"]
        except:
            return "UNKNOWN"

    # ---------------------------------------------------------------------
    def result(self) -> "Mission":
        """Returns the results of the process"""
        if self.status.upper() == "COMPLETED":
            resp = self._con.get(self._url, {"f": "json"})
            if self.properties["type"] == "createNewMissionService":
                url = f"{self._url.split('/jobs/')[0]}/missions/{resp['customAttributes']['missionId']}"
                return Mission(url=url, gis=self._gis)
            else:
                return self.properties
        return None


###########################################################################
class Mission(object):
    """
    A single registered `Mission` on the Enterprise.
    """

    _properties = None
    _url = None
    _con = None
    _gis = None

    # ---------------------------------------------------------------------
    def __init__(self, url: str, gis: "GIS", **kwargs) -> "Mission":
        self._url = url
        self._gis = gis
        self._con = gis._con

    # ---------------------------------------------------------------------
    def _dictionary_check(self, i):
        """
        First prints the final entry in the dictionary (most nested) and its key
        Then prints the keys leading into this
        * could be reversed to be more useful, I guess
        """
        for key, value in i.items():
            if isinstance(value, dict):
                self._dictionary_check(value)
            else:
                if (
                    isinstance(value, str)
                    and value.lower().find("/featureserver/") > -1
                ):
                    i[key] = Service(url=value, server=self._gis)

    # ---------------------------------------------------------------------
    @cached_property
    def properties(self):
        """returns the properties of the resource"""
        if self._properties is None:
            try:
                r = self._con.get(self._url, {"f": "json"})
                self._dictionary_check(r)
                self._properties = r
            except:
                r = self._con.post(self._url, {"f": "json"})
                self._dictionary_check(r)
                self._properties = r
        return self._properties

    # ---------------------------------------------------------------------
    def __str__(self):
        return f"< {self.__class__.__name__} @ {self._url} >"

    # ---------------------------------------------------------------------
    def __repr__(self):
        return f"< {self.__class__.__name__} @ {self._url} >"

    # ---------------------------------------------------------------------
    def delete(self) -> bool:
        """
        Deletes a :class:`~arcgis.gis.mission.api.Mission` from the server.

        :return: Boolean

        """
        params = {"f": "json"}
        url = f"{self._url}/delete"
        res = self._con.post(url, params)
        if res.get("status") or res.get("success"):
            return res.get("status") or res.get("success")
        return res

    # ---------------------------------------------------------------------
    def add_report(
        self,
        title: str,
        service_name: str,
        questions: list[dict],
        source: str = "user",
        description: Optional[str] = None,
        tags: Optional[Union[list, str]] = None,
        display_field: Optional[str] = None,
        drawing_info: Optional[dict] = None,
        locale: str = "en",
        share_as_template: bool = False,
    ) -> dict:
        """

        This method adds a report to the associated mission.


        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        title	               Required String. The name of the report.
        ------------------     --------------------------------------------------------------------
        service_name           Required String. The desired service name for the report.
                               This value cannot be longer than 87 characters and cannot contain the following values:
                               #,$, %,^,*,=,`,{,},[,],,,{space},:,*, ?, ",<,>,|,\, /, +
        ------------------     --------------------------------------------------------------------
        questions              Required String. A string formatted as a JSON Array containing questions and their fields. If an empty array is passed, the request is rejected. Used to represent the desired questions and their fields. Question types are based on Survey123 question type fields.
                               Available question types: Single Line Text, Single Choice, Number,
                               Image, Multiline Text, Dropdown, Multiple Choice, and Date/Time.
                               See https://doc.arcgis.com/en/survey123/browser/create-surveys/quickreferencecreatesurveys.htm#GUID-2D96112F-85B1-4C41-9C6F-A85BB6026A51 for details.
        ------------------     --------------------------------------------------------------------
        description	           Optional String. A description of the report.
        ------------------     --------------------------------------------------------------------
        tags	               Optional. A comma-separated list of strings. Used to add tags to the report.
        ------------------     --------------------------------------------------------------------
        display_field          Optional String. The name of the report feature layer's display field. Normally this is the field name of one of the report questions. If absent, the report feature layer's display field will be the first question's field name.
        ------------------     --------------------------------------------------------------------
        drawing_info           Optional. JSON Object. Defines the report feature layer's drawing info, including a feature renderer.
                               See: https://developers.arcgis.com/documentation/common-data-types/drawinginfo.htm
        ------------------     --------------------------------------------------------------------
        locale                 Optional. String. The locale used to generate the report. Must be a valid IETF BCP 47 language tag. Defaults to en
        ------------------     --------------------------------------------------------------------
        share_as_template      Optional Boolean. Shares the report as a template.
        ==================     ====================================================================

        :return: Dict
        """
        params = {
            "title": title,
            "serviceName": service_name,
            "source": source,
            "questions": questions,
            "description": description or "",
            "tags": tags or "report",
            "displayField": display_field or "",
            "drawingInfo": drawing_info or "",
            "shareAsTemplate": share_as_template,
            "locale": locale,
            "f": "json",
        }
        url = f"{self._url}/addReport"
        return self._con.post(url, params)


###########################################################################
class MissionCatalog:
    """
    The ArcGIS Mission Server catalog.
    """

    _con = None
    _gis = None
    _url = None
    _admin = None
    _properties = None

    # ---------------------------------------------------------------------
    def __init__(self, gis: "GIS") -> "MissionCatalog":
        url = None
        urls = {"admin": None, "url": None}
        for rs in gis._registered_servers()["servers"]:
            if rs["serverType"] == "ARCGIS_MISSION_SERVER":
                url = f"{rs['url']}/rest"
                urls["url"] = url
                urls["admin"] = f"{rs['adminUrl']}/admin"

        if url is None:
            raise Exception("No registered mission server found.")

        self._url = url
        self._gis = gis
        self._con = gis._con
        self._urls = urls

    # ---------------------------------------------------------------------
    def __str__(self):
        return f"< {self.__class__.__name__} @ {self._url} >"

    # ---------------------------------------------------------------------
    def __repr__(self):
        return f"< {self.__class__.__name__} @ {self._url} >"

    # ---------------------------------------------------------------------
    @cached_property
    def properties(self):
        """returns the properties of the resource"""

        if self._properties is None:
            try:
                self._properties = PropertyMap(self._con.get(self._url, {"f": "json"}))
            except:
                self._properties = PropertyMap(self._con.post(self._url, {"f": "json"}))
        return self._properties

    def admin(self):
        """provides a connection to the administrative API of Mission Server Produce"""
        if self._admin is None:
            from arcgis.gis.mission import MissionServer

            self._admin = MissionServer(url=self._urls["admin"], gis=self._gis)
        return self._admin

    # ---------------------------------------------------------------------
    def create_mission(
        self,
        title: str,
        snippet: Optional[str] = None,
        description: Optional[str] = None,
        license_info: Optional[str] = None,
        tags: Optional[Union[list, str]] = None,
        capabilities: Optional[Union[list, str]] = None,
        extent: Optional[str] = None,
        template_item: Optional[Item] = None,
        locale: str = "en",
        base_map: Optional[dict] = None,
        wm_description: Optional[str] = None,
        webmap_id: Optional[Union[str, Item]] = None,
        app_id: Optional[str] = None,
    ) -> MissionJob:
        """

        Creates a new `Mission` on the enterprise.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        title	               Required. String. The title of the mission. This is the human readable title that is displayed to users.
        ------------------     --------------------------------------------------------------------
        snippet	               Optional String. A short summary description of the item.
        ------------------     --------------------------------------------------------------------
        description	           Optional String. Mission description.
        ------------------     --------------------------------------------------------------------
        capabilities	       Optional String. The comma-separated list of desired capabilites for the mission.
                               Valid values are CHAT, TASK, REPORT, PRESENCE.
        ------------------     --------------------------------------------------------------------
        license_info	       Optional String. Any license information or restrictions.
        ------------------     --------------------------------------------------------------------
        tags	               Optional. Comma-separated list of strings used to tag the mission.
                               Format: tag1,tag2,...,tagN
        ------------------     --------------------------------------------------------------------
        extent	               Optional String. Comma-separated list that defines the bounding
                               rectangle of the mission. Should always be in WGS84. The
                               default is -180, -90, 180, 90.

                               **Format: <xmin>, <ymin>, <xmax>, <ymax>**
        ------------------     --------------------------------------------------------------------
        webmap_id	       Optional String. The portal item id of the web map to use as a template for the mission.
        ------------------     --------------------------------------------------------------------
        locale	               Optional String. The locale in which to generate Mission assets with. Must be a valid IETF BCP 47 language tag. Defaults to en
        ------------------     --------------------------------------------------------------------
        base_map               Optional JSON Object. The basemap to add to the mission.
                               See: https://developers.arcgis.com/documentation/common-data-types/basemap.htm
        ------------------     --------------------------------------------------------------------
        wm_description         Optional string. The description of the web map added to the mission.
        ------------------     --------------------------------------------------------------------
        app_id                 Optional string. The ID of the app that created the mission.
        ==================     ====================================================================


        :return:
            :class:`~arcgis.gis.mission.api.MissionJob`


        """
        if isinstance(webmap_id, Item):
            webmap_id = webmap_id.itemid
        url = f"{self._url}/services/createMission"
        params = {
            "title": title,
            "snippet": snippet or "",
            "description": description or "",
            "licenseInfo": license_info or "",
            "capabilities": capabilities or "",
            "async": True,
            "tags": tags or "",
            "extent": extent or "-180,-90,180,90",
            "locale": locale or "en",
            "baseMap": base_map or "",
            "appId": app_id or "",
            "webMapDescription": wm_description,
            "templateWebMapId": webmap_id,
            "f": "json",
        }
        resp = self._con.post(url, params)
        return MissionJob(url=f"{self._url}/jobs/{resp.get('jobId')}", gis=self._gis)

    # ---------------------------------------------------------------------
    @property
    def jobs(self) -> list:
        """returns a list of :class:`~arcgis.gis.mission.api.MissionJob` on the server"""
        url = f"{self._url}/jobs"
        params = {"f": "json"}
        resp = self._con.get(url, params)
        return [MissionJob(url=f"{url}/{j}", gis=self._gis) for j in resp["asyncJobs"]]

    # ---------------------------------------------------------------------
    @property
    def missions(self) -> list:
        """
        returns a list of :class:`~arcgis.gis.mission.api.Mission` on the server

        :return: List
        """
        url = f"{self._url}/services"
        params = {"f": "json"}
        resp = self._con.get(url, params)
        return [
            Mission(url=f"{url}/{j['name']}/MissionServer", gis=self._gis)
            for j in resp["services"]
            if j["type"] == "MissionServer"
        ]



from __future__ import annotations
import logging
from enum import Enum
from typing import Any
from functools import lru_cache
from arcgis.auth.tools import LazyLoader

_arcgis = LazyLoader("arcgis")
requests = LazyLoader("requests")
_dt = LazyLoader("datetime")

logger = logging.getLogger(__name__)

__all__ = [
    "DataStoreMetric",
    "DataStoreTimeUnit",
    "DataStoreAggregation",
    "DataStoreMetricsManager",
]


class DataStoreMetric(Enum):
    """The allowed metric values"""

    # Average CPU Time
    AVG_CPU = "db_avg_cpu"
    # Storage Size in MegaBytes
    STORAGE_MB = "db_used_storage_mb"
    # Allocated Storage Size in MegaBytes
    ALLOCATED_STORAGE = "db_allocated_storage_mb"
    # Percent of Total Storage
    PERCENT_STORAGE = "db_percent_storage"
    # Size of Feature Storage in MegaBytes
    FEATURESTORAGE = "db_featurestorage_mb"


class DataStoreTimeUnit(Enum):
    """Time Units for DataStore Metrics class"""

    # a period of twenty-four hours as a unit of time
    DAY = "d"
    # a period of time equal to a twenty-fourth part of a day and night and divided into 60 minutes.
    HOUR = "h"
    # a period of time equal to sixty seconds or a sixtieth of an hour.
    MINUTE = "m"
    #  a unit of time in the International System of Units (SI), historically defined as 1/86400 of a day
    SECOND = "s"
    # a period of time equal to one thousandth of a second.
    MILLISECOND = "ms"


class DataStoreAggregation(Enum):
    # a number expressing the central or typical value in a set of data
    AVG = "avg"
    # determine the total number of (a collection of items).
    COUNT = "count"
    # a maximum amount
    MAX = "max"
    # a minimum amount
    MIN = "min"
    # a quantity calculated to indicate the extent of deviation for a group as a whole.
    STDEV = "stdev"
    # the total amount resulting from the addition of two or more numbers, amounts, or items.
    SUM = "sum"
    # a statistical measurement of the spread between numbers in a data set
    VARIANCE = "variance"


class DataStoreMetricsManager:
    """
    This class allows for ArcGIS Online administrators to query statistics about the
    managed datastore. It is not meant to be initialized directly, but instead an
    instance is returned from the :attr:`~arcgis.gis.admin.AGOLAdminManager.datastore_metrics`
    property.

    .. code-block:: python

        # Usage Example;
        >>> gis = GIS(profile="your_online_admin_profile")

        >>> ago_mgr = gis.admin
        >>> ds_mgr = ago_mgr.datastore_metrics
    """

    _gis: _arcgis.gis.GIS | None = None
    _orgid: str | None = None
    _session: _arcgis.auth.EsriSession | None = None
    _base_url: str = None

    # ---------------------------------------------------------------------
    def __init__(self, gis: _arcgis.gis.GIS):
        self._gis = gis
        self._session = gis._con._session
        self._orgid = gis.properties.id
        self._base_url = gis.properties["helperServices"]["datastoreManagement"]["url"]

    # ---------------------------------------------------------------------
    def __str__(self) -> str:
        return "< Datastore Metrics >"

    # ---------------------------------------------------------------------
    def __repr__(self) -> str:
        return self.__str__()

    # ---------------------------------------------------------------------
    def query_resource_usage(self, query_period: str) -> dict[str, list]:
        """
        Allows administrators to view the average and max CPU usage for the
        organization's datastore.

        ====================  =========================================================
        **Parameter**         **Description**
        --------------------  ---------------------------------------------------------
        query_period          Required String. The time period to query. The allowed
                              values are: day, week, or hour.
        ====================  =========================================================

        :return:
           dictionary containing the average and max usage.
        """
        query_period = query_period.lower()
        assert query_period in [
            "day",
            "week",
            "hour",
        ], "`query_period` must be a value of `day`, `week` or `hour`."

        unit_lu: dict[str, str] = {
            "day": [
                {
                    "metric": DataStoreMetric.AVG_CPU.value,
                    "binAggregation": DataStoreAggregation.AVG.value,
                    "binUnit": DataStoreTimeUnit.MINUTE.value,
                    "agoUnit": DataStoreTimeUnit.DAY.value,
                    "ago": 1,
                    "binSize": 5,
                },
                {
                    "metric": DataStoreMetric.AVG_CPU.value,
                    "binAggregation": DataStoreAggregation.MAX.value,
                    "binUnit": DataStoreTimeUnit.MINUTE.value,
                    "agoUnit": DataStoreTimeUnit.DAY.value,
                    "ago": 1,
                    "binSize": 5,
                },
            ],
            "week": [
                {
                    "metric": DataStoreMetric.AVG_CPU.value,
                    "binAggregation": DataStoreAggregation.AVG.value,
                    "binUnit": DataStoreTimeUnit.MINUTE.value,
                    "agoUnit": DataStoreTimeUnit.DAY.value,
                    "ago": 7,
                    "binSize": 30,
                },
                {
                    "metric": DataStoreMetric.AVG_CPU.value,
                    "binAggregation": DataStoreAggregation.MAX.value,
                    "binUnit": DataStoreTimeUnit.MINUTE.value,
                    "agoUnit": DataStoreTimeUnit.DAY.value,
                    "ago": 7,
                    "binSize": 30,
                },
            ],
            "hour": [
                {
                    "metric": DataStoreMetric.AVG_CPU.value,
                    "binAggregation": DataStoreAggregation.AVG.value,
                    "binUnit": DataStoreTimeUnit.MINUTE.value,
                    "agoUnit": DataStoreTimeUnit.HOUR.value,
                    "ago": 1,
                    "binSize": 1,
                },
                {
                    "metric": DataStoreMetric.AVG_CPU.value,
                    "binAggregation": DataStoreAggregation.MAX.value,
                    "binUnit": DataStoreTimeUnit.MINUTE.value,
                    "agoUnit": DataStoreTimeUnit.HOUR.value,
                    "ago": 1,
                    "binSize": 1,
                },
            ],
        }
        params: dict[str, Any] = {
            "f": "json",
        }
        url: str = f"{self._base_url}/{self._orgid}/datastores/featureDataStore"

        return_data: dict = {}
        for param in unit_lu[query_period]:
            params.update(param)
            resp: requests.Response = self._session.get(url=url, params=params)
            resp.raise_for_status()
            data: list = resp.json()
            if params["binAggregation"] == DataStoreAggregation.MAX.value:
                return_data["max"] = [
                    {
                        "ts": _dt.datetime.fromtimestamp(entry["ts"] / 1000.0),
                        "value": entry["value"],
                    }
                    for entry in data
                ]
            else:
                return_data["avg"] = [
                    {
                        "ts": _dt.datetime.fromtimestamp(entry["ts"] / 1000.0),
                        "value": entry["value"],
                    }
                    for entry in data
                ]
        return return_data

    # ---------------------------------------------------------------------
    @property
    def feature_storage(self) -> list[dict[str, Any]]:
        """
        Returns storage percentage of total storage used

        :returns: list[dict[str,Any]]

        """
        params: dict[str, Any] = {
            "f": "json",
            "metric": DataStoreMetric.FEATURESTORAGE.value,
            "ago": 1,
            "agoUnit": DataStoreTimeUnit.HOUR.value,
        }
        url: str = f"{self._base_url}/{self._orgid}/datastores/featureDataStore"
        resp: requests.Response = self._session.get(url=url, params=params)
        resp.raise_for_status()
        data: list[dict[str, Any]] = resp.json()
        return [
            {
                "ts": _dt.datetime.fromtimestamp(entry["ts"] / 1000.0),
                "value": int(round(entry["value"] / 500000, 2) * 100),
            }
            for entry in data
        ]

    # ---------------------------------------------------------------------
    @lru_cache(maxsize=255)
    def query(
        self,
        metric: DataStoreMetric,
        bin_size: float | int,
        bin_unit: DataStoreTimeUnit,
        aggregation: DataStoreAggregation = DataStoreAggregation.SUM,
        start_time: _dt.datetime | None = None,
        end_time: _dt.datetime | None = None,
        ago: int | None = None,
        ago_unit: DataStoreTimeUnit | None = None,
    ) -> list[dict[str, Any]]:
        """
        A query operation used to gather metrics about the ArcGIS Online Datastore

        ====================  =========================================================
        **Parameter**         **Description**
        --------------------  ---------------------------------------------------------
        metric                Required DataStoreMetric. The statistical method to gather.
        --------------------  ---------------------------------------------------------
        bin_size              Required Float or Int. The size of the bin to aggregate on.
        --------------------  ---------------------------------------------------------
        bin_unit              Required DataStoreTimeUnit. The size of the bin.
        --------------------  ---------------------------------------------------------
        aggregation           Required DataStoreAggregation. The type of aggregation to perform.
        --------------------  ---------------------------------------------------------
        start_time            Optional datetime.datetime. The starting date point.
        --------------------  ---------------------------------------------------------
        end_time              Optional datetime.datetime. The ending date point.
        --------------------  ---------------------------------------------------------
        ago                   Optional Int. The time to look back from today.
        --------------------  ---------------------------------------------------------
        ago_unit              Optional DataStoreTimeUnit. The time unit to look back.
        ====================  =========================================================

        :returns: list[dict[str,Any]]
        """
        if ago is None and start_time is None and end_time is None:
            logger.warning("No ago was supplied, defaulting to ago=1")
            ago = 1
            ago_unit = DataStoreTimeUnit.DAY
        params: dict[str, Any] = {
            "f": "json",
        }
        assert isinstance(metric, DataStoreMetric)
        params["metric"] = metric.value
        if start_time:
            assert isinstance(
                start_time, _dt.datetime
            ), "`start_time` must be a datetime object"
            params["startMillis"] = int(start_time.timestamp() * 1000)
        if end_time:
            assert isinstance(
                end_time, _dt.datetime
            ), "`end_time` must be a datetime object"
            params["endMillis"] = int(end_time.timestamp() * 1000)
        if ago:
            assert isinstance(ago, int), "`ags` must be an integer."
            params["ago"] = ago
        if ago_unit:
            assert isinstance(
                ago_unit, DataStoreTimeUnit
            ), "`ago_unit` must be a DataStoreTimeUnit"
            params["agoUnit"] = ago_unit.value
        if bin_unit:
            bin_unit
            assert isinstance(
                bin_unit, DataStoreTimeUnit
            ), "`ago_unit` must be a DataStoreTimeUnit"
            params["binUnit"] = bin_unit.value
        if bin_size:
            assert isinstance(
                bin_size, (int, float)
            ), "`ago_unit` must be a float or int."
            params["binSize"] = bin_size
        if aggregation:
            assert isinstance(
                aggregation, DataStoreAggregation
            ), "`aggregation` must be a float or int."
            params["binAggregation"] = aggregation.value
        if start_time and end_time and ago:
            logger.warning(
                "Ignoring the `ago` parameter due to the presents of start and end times."
            )
            del params["ago"]
        url: str = f"{self._base_url}/{self._orgid}/datastores/featureDataStore"
        resp: requests.Response = self._session.get(url=url, params=params)
        resp.raise_for_status()
        data: list[dict[str, Any]] = resp.json()
        return [
            {
                "ts": _dt.datetime.fromtimestamp(entry["ts"] / 1000.0),
                "value": entry["value"],
            }
            for entry in data
        ]



from requests.auth import AuthBase
from requests import PreparedRequest, Response


###########################################################################
class BaseEsriAuth(AuthBase):
    """
    A base class that developers can inherit from to create custom
    authentication handlers.

    For more information, please see:
    https://docs.python-requests.org/en/master/user/authentication/#new-forms-of-authentication
    """

    # ---------------------------------------------------------------------
    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        """
        Called Before PrepareRequest is Completed.

        The logic to attached preemptive authentication should go here.
        If you want to wait for response, register a response_hook within the call.

        :returns: PrepareRequest

        ```
        def __call__(self, r:PreparedRequest) -> PreparedRequest:
            r.register_hook(event="response", hook=self.response_hook)
            return r
        ```

        """
        raise NotImplementedError("Auth hooks must be callable.")

    # ---------------------------------------------------------------------
    def response_hook(self, response: Response, **kwargs) -> Response:
        """
        response hook logic

        :return: Response
        """
        return response

    # ---------------------------------------------------------------------
    @property
    def token(self) -> str:
        """
        returns a ArcGIS Token as a string

        :return: string
        """
        raise NotImplementedError("Token not implemented.")



"Functions for calling the Deep Learning Tools."
from . import _utils
from ._utils.env import _LAMBDA_TEXT_CLASSIFICATION
from arcgis.geoprocessing._support import (
    _analysis_job,
    _analysis_job_results,
    _analysis_job_status,
    _layer_input,
)
import json as _json
import arcgis as _arcgis
from arcgis.raster._layer import ImageryLayer as _ImageryLayer
from arcgis.raster._util import _set_context, _id_generator
from ._scannedmapdigitizer import ScannedMapDigitizer
from .models._timm_utils import load_timm_bckbn_pretrained
from timm.models import helpers

helpers.load_pretrained = load_timm_bckbn_pretrained

if not _LAMBDA_TEXT_CLASSIFICATION:
    from .models import (
        SingleShotDetector,
        UnetClassifier,
        FeatureClassifier,
        RetinaNet,
        PSPNetClassifier,
        MaskRCNN,
        DeepLab,
        PointCNN,
        ModelExtension,
        FasterRCNN,
        SuperResolution,
        FullyConnectedNetwork,
        MLModel,
        YOLOv3,
        HEDEdgeDetector,
        BDCNEdgeDetector,
        ImageCaptioner,
        TimeSeriesModel,
        CycleGAN,
        MultiTaskRoadExtractor,
        ChangeDetector,
        Pix2Pix,
        ConnectNet,
        SiamMask,
        Track,
        Embeddings,
        MMDetection,
        MMSegmentation,
        AutoML,
        DeepSort,
        Pix2PixHD,
        AutoDL,
        ImageryModel,
        MaXDeepLab,
        WNet_cGAN,
        DETReg,
        RandLANet,
        EfficientDet,
        SQNSeg,
        PSETAE,
        MMDetection3D,
    )

    from ._object_tracker import ObjectTracker

    from ._utils.pointcloud_data import Transform3d
from ._data import prepare_data, prepare_tabulardata, prepare_textdata
from ._process_df import process_df, add_datepart
from ._utils.evaluate_batchsize import estimate_batch_size


_point_cloud_classification_model_list = ["PointCNN", "RandLANet", "SQNSeg"]


_point_cloud_detection_model_list = ["MMDetection3D"]


def _set_param(gis, params, param_name, input_param):
    if isinstance(input_param, str):
        if "http:" in input_param or "https:" in input_param:
            params[param_name] = _json.dumps({"url": input_param})
        else:
            params[param_name] = _json.dumps({"uri": input_param})

    elif isinstance(input_param, _arcgis.gis.Item):
        params[param_name] = _json.dumps({"itemId": input_param.itemid})

    elif isinstance(input_param, dict):
        params[param_name] = input_param
    elif isinstance(input_param, Model):
        params[param_name] = input_param._model
    else:
        raise TypeError(input_param + " should be a string (service url) or Item")

    return


def _create_output_image_service(gis, output_name, task, folder=None):
    ok = gis.content.is_service_name_available(output_name, "Image Service")
    if not ok:
        raise RuntimeError(
            "An Image Service by this name already exists: " + output_name
        )

    create_parameters = {
        "name": output_name,
        "description": "",
        "capabilities": "Image",
        "properties": {"path": "@", "description": "", "copyright": ""},
    }

    output_service = gis.content.create_service(
        output_name,
        create_params=create_parameters,
        service_type="imageService",
        folder=folder,
    )
    description = "Image Service generated from running the " + task + " tool."
    item_properties = {
        "description": description,
        "tags": "Analysis Result, " + task,
        "snippet": "Analysis Image Service generated from " + task,
    }
    output_service.update(item_properties)
    return output_service


def _create_output_feature_service(
    gis,
    output_name,
    output_service_name="Analysis feature service",
    task="GeoAnalytics",
    folder=None,
):
    ok = gis.content.is_service_name_available(output_name, "Feature Service")
    if not ok:
        raise RuntimeError(
            "A Feature Service by this name already exists: " + output_name
        )

    createParameters = {
        "currentVersion": 10.2,
        "serviceDescription": "",
        "hasVersionedData": False,
        "supportsDisconnectedEditing": False,
        "hasStaticData": True,
        "maxRecordCount": 2000,
        "supportedQueryFormats": "JSON",
        "capabilities": "Query",
        "description": "",
        "copyrightText": "",
        "allowGeometryUpdates": False,
        "syncEnabled": False,
        "editorTrackingInfo": {
            "enableEditorTracking": False,
            "enableOwnershipAccessControl": False,
            "allowOthersToUpdate": True,
            "allowOthersToDelete": True,
        },
        "xssPreventionInfo": {
            "xssPreventionEnabled": True,
            "xssPreventionRule": "InputOnly",
            "xssInputRule": "rejectInvalid",
        },
        "tables": [],
        "name": output_service_name.replace(" ", "_"),
    }

    output_service = gis.content.create_service(
        output_name,
        create_params=createParameters,
        service_type="featureService",
        folder=folder,
    )
    description = "Feature Service generated from running the " + task + " tool."
    item_properties = {
        "description": description,
        "tags": "Analysis Result, " + task,
        "snippet": output_service_name,
    }
    output_service.update(item_properties)
    return output_service


def _set_output_raster(output_name, task, gis, output_properties=None):
    output_service = None
    output_raster = None

    task_name = task

    folder = None
    folderId = None

    if output_properties is not None:
        if "folder" in output_properties:
            folder = output_properties["folder"]
    if folder is not None:
        if isinstance(folder, dict):
            if "id" in folder:
                folderId = folder["id"]
                folder = folder["title"]
        else:
            owner = gis.properties.user.username
            folderId = gis._portal.get_folder_id(owner, folder)
        if folderId is None:
            folder_dict = gis.content.create_folder(folder, owner)
            folder = folder_dict["title"]
            folderId = folder_dict["id"]

    if output_name is None:
        output_name = str(task_name) + "_" + _id_generator()
        output_service = _create_output_image_service(
            gis, output_name, task, folder=folder
        )
        output_raster = {
            "serviceProperties": {
                "name": output_service.name,
                "serviceUrl": output_service.url,
            },
            "itemProperties": {"itemId": output_service.itemid},
        }
    elif isinstance(output_name, str):
        output_service = _create_output_image_service(
            gis, output_name, task, folder=folder
        )
        output_raster = {
            "serviceProperties": {
                "name": output_service.name,
                "serviceUrl": output_service.url,
            },
            "itemProperties": {"itemId": output_service.itemid},
        }
    elif isinstance(output_name, _arcgis.gis.Item):
        output_service = output_name
        output_raster = {"itemProperties": {"itemId": output_service.itemid}}
    else:
        raise TypeError("output_raster should be a string (service name) or Item")

    if folderId is not None:
        output_raster["itemProperties"].update({"folderId": folderId})
    output_raster = _json.dumps(output_raster)
    return output_raster, output_service


def detect_objects(
    input_raster,
    model,
    model_arguments=None,
    output_name=None,
    run_nms=False,
    confidence_score_field=None,
    class_value_field=None,
    max_overlap_ratio=0,
    context=None,
    process_all_raster_items=False,
    *,
    gis=None,
    future=False,
    **kwargs,
):
    """
    Function can be used to generate feature service that contains polygons on detected objects
    found in the imagery data using the designated deep learning model. Note that the deep learning
    library needs to be installed separately, in addition to the server's built in Python 3.x library.

    .. note::
            This function is supported with ArcGIS Enterprise (Image Server) and ArcGIS Image for ArcGIS Online.

    ====================================     ====================================================================
    **Parameter**                             **Description**
    ------------------------------------     --------------------------------------------------------------------
    input_raster                             Required. raster layer that contains objects that needs to be detected.
    ------------------------------------     --------------------------------------------------------------------
    model                                    Required :class:`~arcgis.learn.Model` object.
    ------------------------------------     --------------------------------------------------------------------
    model_arguments                          Optional dictionary. Name-value pairs of arguments and their values that can be customized by the clients.

                                             eg: {"name1":"value1", "name2": "value2"}
    ------------------------------------     --------------------------------------------------------------------
    output_name                              Optional. If not provided, a :class:`~arcgis.features.FeatureLayer` is created by the method and used as the output .
                                             You can pass in an existing Feature Service Item from your GIS to use that instead.
                                             Alternatively, you can pass in the name of the output Feature Service that should be created by this method
                                             to be used as the output for the tool.
                                             A RuntimeError is raised if a service by that name already exists
    ------------------------------------     --------------------------------------------------------------------
    run_nms                                  Optional bool. Default value is False. If set to True, runs the Non Maximum Suppression tool.
    ------------------------------------     --------------------------------------------------------------------
    confidence_score_field                   Optional string. The field in the feature class that contains the confidence scores as output by the object detection method.
                                             This parameter is required when you set the run_nms to True
    ------------------------------------     --------------------------------------------------------------------
    class_value_field                        Optional string. The class value field in the input feature class.
                                             If not specified, the function will use the standard class value fields
                                             Classvalue and Value. If these fields do not exist, all features will
                                             be treated as the same object class.
                                             Set only if run_nms  is set to True
    ------------------------------------     --------------------------------------------------------------------
    max_overlap_ratio                        Optional integer. The maximum overlap ratio for two overlapping features.
                                             Defined as the ratio of intersection area over union area.
                                             Set only if run_nms  is set to True
    ------------------------------------     --------------------------------------------------------------------
    context                                  Optional dictionary. Context contains additional settings that affect task execution.
                                             Dictionary can contain value for following keys:

                                             - cellSize - Set the output raster cell size, or resolution

                                             - extent - Sets the processing extent used by the function

                                             - parallelProcessingFactor - Sets the parallel processing factor. Default is "80%"

                                             - mask: Only cells that fall within the analysis mask will be considered in the operation.

                                             Eg: {"mask": {"url": "<feature_service_url>"}}

                                             - processorType - Sets the processor type. "CPU" or "GPU"

                                             Eg: {"processorType" : "CPU"}

                                             Setting context parameter will override the values set using arcgis.env
                                             variable for this particular function.
    ------------------------------------     --------------------------------------------------------------------
    process_all_raster_items                 Optional bool. Specifies how all raster items in an image service will be processed.

                                             - False : all raster items in the image service will be mosaicked together and processed. This is the default.

                                             - True : all raster items in the image service will be processed as separate images.
    ------------------------------------     --------------------------------------------------------------------
    gis                                      Optional :class:`~arcgis.gis.GIS` . The GIS on which this tool runs. If not specified, the active GIS is used.
    ------------------------------------     --------------------------------------------------------------------
    future                                   Keyword only parameter. Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    ====================================     ====================================================================

    :return:
        The output feature layer item containing the detected objects

    """

    # task = "DetectObjectsUsingDeepLearning"

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.rasteranalysis.detect_objects_using_deep_learning(
        input_raster=input_raster,
        model=model,
        output_objects=output_name,
        model_arguments=model_arguments,
        run_nms=run_nms,
        confidence_score_field=confidence_score_field,
        class_value_field=class_value_field,
        max_overlap_ratio=max_overlap_ratio,
        context=context,
        process_all_raster_items=process_all_raster_items,
        future=future,
        **kwargs,
    )

    """
    url = gis.properties.helperServices.rasterAnalytics.url
    gptool = _arcgis.gis._GISResource(url, gis)

    params = {}

    params["inputRaster"] = _layer_input(input_raster)

    if output_name is None:
        output_service_name = 'DetectObjectsUsingDeepLearning_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    folder = None
    folderId = None
    if kwargs is not None:
        if "folder" in kwargs:
                folder = kwargs["folder"]
        if folder is not None:
            if isinstance(folder, dict):
                if "id" in folder:
                    folderId = folder["id"]
                    folder=folder["title"]
            else:
                owner = gis.properties.user.username
                folderId = gis._portal.get_folder_id(owner, folder)
            if folderId is None:
                folder_dict = gis.content.create_folder(folder, owner)
                folder = folder_dict["title"]
                folderId = folder_dict["id"]

    output_service = _create_output_feature_service(gis, output_name, output_service_name, 'Detect Objects', folder)

    if folderId is not None:
        params["outputObjects"] = _json.dumps({"serviceProperties": {"name": output_service_name, "serviceUrl": output_service.url},
                                               "itemProperties": {"itemId": output_service.itemid}, "folderId":folderId})
    else:
        params["outputObjects"] = _json.dumps({"serviceProperties": {"name": output_service_name, "serviceUrl": output_service.url},
                                               "itemProperties": {"itemId": output_service.itemid}})

    if model is None:
        raise RuntimeError('model cannot be None')
    else:
        _set_param(gis, params, "model", model)

    if model_arguments:
        params["modelArguments"] = dict((str(k),str(v)) for k, v in model_arguments.items())

    if isinstance(run_nms, bool):
        if run_nms:
            params["runNMS"] = True

            if confidence_score_field is not None:
                params["confidenceScoreField"] = confidence_score_field

            if class_value_field is not None:
                params["classValueField"] = class_value_field
    
            if max_overlap_ratio is not None:
                params["maxOverlapRatio"] = max_overlap_ratio
        else:
            params["runNMS"] = False
    else:
        raise RuntimeError("run_nms value should be an instance of bool")
    
    _set_context(params, context)

    task_url, job_info, job_id = _analysis_job(gptool, task, params)

    job_info = _analysis_job_status(gptool, task_url, job_info)
    job_values = _analysis_job_results(gptool, task_url, job_info, job_id)
    item_properties = {
        "properties": {
            "jobUrl": task_url + '/jobs/' + job_info['jobId'],
            "jobType": "GPServer",
            "jobId": job_info['jobId'],
            "jobStatus": "completed"
        }
    }
    output_service.update(item_properties)
    return output_service
    """


def classify_pixels(
    input_raster,
    model,
    model_arguments=None,
    output_name=None,
    context=None,
    process_all_raster_items=False,
    *,
    gis=None,
    future=False,
    **kwargs,
):
    """
    Function to classify input imagery data using a deep learning model.
    Note that the deep learning library needs to be installed separately,
    in addition to the server's built in Python 3.x library.

    .. note::
            This function is supported with ArcGIS Enterprise (Image Server) and ArcGIS Image for ArcGIS Online.

    ====================================     ====================================================================
    **Parameter**                             **Description**
    ------------------------------------     --------------------------------------------------------------------
    input_raster                             Required. raster layer that needs to be classified.
    ------------------------------------     --------------------------------------------------------------------
    model                                    Required :class:`~arcgis.learn.Model` object.
    ------------------------------------     --------------------------------------------------------------------
    model_arguments                          Optional dictionary. Name-value pairs of arguments and their values that can be customized by the clients.

                                             eg: {"name1":"value1", "name2": "value2"}

    ------------------------------------     --------------------------------------------------------------------
    output_name                              Optional. If not provided, an imagery layer is created by the method and used as the output .
                                             You can pass in an existing Image Service Item from your GIS to use that instead.
                                             Alternatively, you can pass in the name of the output Image Service that should be created by this method
                                             to be used as the output for the tool.
                                             A RuntimeError is raised if a service by that name already exists
    ------------------------------------     --------------------------------------------------------------------
    context                                  Optional dictionary. Context contains additional settings that affect task execution.
                                             Dictionary can contain value for following keys:

                                             - outSR - (Output Spatial Reference) Saves the result in the specified spatial reference

                                             - snapRaster - Function will adjust the extent of output rasters so that they
                                               match the cell alignment of the specified snap raster.

                                             - cellSize - Set the output raster cell size, or resolution

                                             - extent - Sets the processing extent used by the function

                                             - parallelProcessingFactor - Sets the parallel processing factor. Default is "80%"

                                             - processorType - Sets the processor type. "CPU" or "GPU"

                                               Example:
                                                    {"outSR" : {spatial reference}}

                                             Setting context parameter will override the values set using arcgis.env
                                             variable for this particular function.
    ------------------------------------     --------------------------------------------------------------------
    process_all_raster_items                 Optional bool. Specifies how all raster items in an image service will be processed.

                                             - False : all raster items in the image service will be mosaicked together and processed. This is the default.

                                             - True : all raster items in the image service will be processed as separate images.
    ------------------------------------     --------------------------------------------------------------------
    gis                                      Optional :class:`~arcgis.gis.GIS` . The GIS on which this tool runs. If not specified, the active GIS is used.
    ------------------------------------     --------------------------------------------------------------------
    future                                   Keyword only parameter. Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    ------------------------------------     --------------------------------------------------------------------
    tiles_only                               Keyword only parameter. Optional boolean.
                                             In ArcGIS Online, the default output image service for this function would be a Tiled Imagery Layer.
                                             To create Dynamic Imagery Layer as output in ArcGIS Online, set tiles_only parameter to False.

                                             Function will not honor tiles_only parameter in ArcGIS Enterprise and will generate Dynamic Imagery Layer by default.
    ====================================     ====================================================================

    :return:
        The classified imagery layer item

    """

    # task = "ClassifyPixelsUsingDeepLearning"

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.rasteranalysis.classify_pixels_using_deep_learning(
        input_raster=input_raster,
        model=model,
        model_arguments=model_arguments,
        output_classified_raster=output_name,
        context=context,
        process_all_raster_items=process_all_raster_items,
        future=future,
        **kwargs,
    )

    """
    url = gis.properties.helperServices.rasterAnalytics.url
    gptool = _arcgis.gis._GISResource(url, gis)

    output_service = None

    output_raster, output_service = _set_output_raster(output_name, task, gis, kwargs)

    params = {}

    params["outputClassifiedRaster"] = output_raster

    params["inputRaster"] = _layer_input(input_raster)

    if model is None:
        raise RuntimeError('model cannot be None')
    else:
        _set_param(gis, params, "model", model)

    if model_arguments:
        params["modelArguments"] = dict((str(k),str(v)) for k, v in model_arguments.items())

    _set_context(params, context)

    task_url, job_info, job_id = _analysis_job(gptool, task, params)

    job_info = _analysis_job_status(gptool, task_url, job_info)
    job_values = _analysis_job_results(gptool, task_url, job_info, job_id)
    item_properties = {
        "properties": {
            "jobUrl": task_url + '/jobs/' + job_info['jobId'],
            "jobType": "GPServer",
            "jobId": job_info['jobId'],
            "jobStatus": "completed"
        }
    }
    output_service.update(item_properties)
    return output_service
    """


def export_training_data(
    input_raster,
    input_class_data=None,
    chip_format=None,
    tile_size=None,
    stride_size=None,
    metadata_format=None,
    classvalue_field=None,
    buffer_radius=None,
    output_location=None,
    context=None,
    input_mask_polygons=None,
    rotation_angle=0,
    reference_system="MAP_SPACE",
    process_all_raster_items=False,
    blacken_around_feature=False,
    fix_chip_size=True,
    additional_input_raster=None,
    input_instance_data=None,
    instance_class_value_field=None,
    min_polygon_overlap_ratio=0,
    *,
    gis=None,
    future=False,
    **kwargs,
):
    """
    Function is designed to generate training sample image chips from the input imagery data with
    labeled vector data or classified images. The output of this service tool is the data store string
    where the output image chips, labels and metadata files are going to be stored.

    .. note::
            This function is supported with ArcGIS Enterprise (Image Server)

    ====================================     ====================================================================
    **Parameter**                             **Description**
    ------------------------------------     --------------------------------------------------------------------
    input_raster                             Required :class:`~arcgis.raster.ImageryLayer`/:class:`~arcgis.raster.Raster`/:class:`~arcgis.gis.Item`/String (URL).
                                             Raster layer that needs to be exported for training.
    ------------------------------------     --------------------------------------------------------------------
    input_class_data                         Labeled data, either a feature layer or image layer.
                                             Vector inputs should follow a training sample format as
                                             generated by the ArcGIS Pro Training Sample Manager.
                                             Raster inputs should follow a classified raster format as generated by the Classify Raster tool.
    ------------------------------------     --------------------------------------------------------------------
    chip_format                              Optional string. The raster format for the image chip outputs.

                                             - ``TIFF``: TIFF format

                                             - ``PNG``: PNG format

                                             - ``JPEG``: JPEG format

                                             - ``MRF``: MRF (Meta Raster Format)
    ------------------------------------     --------------------------------------------------------------------
    tile_size                                Optional dictionary. The size of the image chips.

                                             Example: {"x": 256, "y": 256}
    ------------------------------------     --------------------------------------------------------------------
    stride_size                              Optional dictionary. The distance to move in the X and Y when creating
                                             the next image chip.
                                             When stride is equal to the tile size, there will be no overlap.
                                             When stride is equal to half of the tile size, there will be 50% overlap.

                                             Example: {"x": 128, "y": 128}
    ------------------------------------     --------------------------------------------------------------------
    metadata_format                          Optional string. The format of the output metadata labels. There are 4 options for output metadata labels for the training data,
                                             KITTI Rectangles, PASCAL VOCrectangles, Classified Tiles (a class map) and RCNN_Masks. If your input training sample data
                                             is a feature class layer such as building layer or standard classification training sample file,
                                             use the KITTI or PASCAL VOC rectangle option.
                                             The output metadata is a .txt file or .xml file containing the training sample data contained
                                             in the minimum bounding rectangle. The name of the metadata file matches the input source image
                                             name. If your input training sample data is a class map, use the Classified Tiles as your output metadata format option.

                                             - ``KITTI_rectangles``: The metadata follows the same format as the Karlsruhe Institute of Technology and Toyota echnological Institute (KITTI) Object Detection Evaluation dataset. The KITTI dataset is a vision benchmark suite. This is the default.The label files are plain text files. All values, both numerical or strings, are separated by spaces, and each row corresponds to one object. This format can be used with FasterRCNN, RetinaNet, SingleShotDetector and YOLOv3 models.

                                             - ``PASCAL_VOC_rectangles``: The metadata follows the same format as the Pattern Analysis, Statistical Modeling and
                                               Computational Learning, Visual Object Classes (PASCAL_VOC) dataset. The PASCAL VOC dataset is a standardized
                                               image data set for object class recognition.The label files are XML files and contain information about image name,
                                               class value, and bounding box(es).
                                               This format can be used with FasterRCNN, RetinaNet, SingleShotDetector and YOLOv3 models.

                                             - ``Classified_Tiles``: This option will output one classified image chip per input image chip.
                                               No other meta data for each image chip. Only the statistics output has more information on the
                                               classes such as class names, class values, and output statistics.
                                               This format can be used with BDCNEdgeDetector, DeepLab, HEDEdgeDetector, MultiTaskRoadExtractor, PSPNetClassifier and UnetClassifier models.

                                             - ``RCNN_Masks``: This option will output image chips that have a mask on the areas where the sample exists.
                                               The model generates bounding boxes and segmentation masks for each instance of an object in the image.
                                               This format can be used with MaskRCNN model.

                                             - ``Labeled_Tiles``: This option will label each output tile with a specific class.
                                               This format is used for image classification.
                                               This format can be used with FeatureClassifier model.

                                             - ``MultiLabeled_Tiles``: Each output tile will be labeled with one or more classes.
                                               For example, a tile may be labeled agriculture and also cloudy. This format is used for object classification.
                                               This format can be used with FeatureClassifier model.

                                             - ``Export_Tiles``: The output will be image chips with no label.
                                               This format is used for image enhancement techniques such as Super Resolution and Change Detection.
                                               This format can be used with ChangeDetector, CycleGAN, Pix2Pix and SuperResolution models.

                                             - ``CycleGAN``: The output will be image chips with no label. This format is used for image
                                               translation technique CycleGAN, which is used to train images that do not overlap.

                                             - ``Imagenet``: Each output tile will be labeled with a specific class. This format is used
                                               for object classification; however, it can also be used for object tracking when the Deep Sort
                                               model type is used during training.

                                             - ``Panoptic_Segmentation``: The output will be one classified image chip and one instance per
                                               input image chip. The output will also have image chips that mask the areas where the sample exists;
                                               these image chips will be stored in a different folder. This format is used for both pixel classification
                                               and instance segmentation, therefore there will be two output labels folders.
    ------------------------------------     --------------------------------------------------------------------
    classvalue_field                         Optional string. Specifies the field which contains the class values. If no field is specified,
                                             the system will look for a 'value' or 'classvalue' field. If this feature does
                                             not contain a class field, the system will presume all records belong the 1 class.
    ------------------------------------     --------------------------------------------------------------------
    buffer_radius                            Optional integer. Specifies a radius for point feature classes to specify training sample area.
    ------------------------------------     --------------------------------------------------------------------
    output_location                          This is the output location for training sample data.
                                             It can be the server data store path or a shared file system path.

                                             Example:

                                             Server datastore path -

                                                * ``/fileShares/deeplearning/rooftoptrainingsamples``
                                                * ``/rasterStores/rasterstorename/rooftoptrainingsamples``

                                             File share path -

                                                * ``\\\\servername\\deeplearning\\rooftoptrainingsamples``
    ------------------------------------     --------------------------------------------------------------------
    context                                  Optional dictionary. Context contains additional settings that affect task execution.
                                             Dictionary can contain value for following keys:

                                             - exportAllTiles - Choose if the image chips with overlapped labeled data will be exported.

                                               * True - Export all the image chips, including those that do not overlap labeled data.
                                               * False - Export only the image chips that overlap the labelled data. This is the default.

                                             - startIndex - Allows you to set the start index for the sequence of image chips.
                                               This lets you append more image chips to an existing sequence. The default value is 0.

                                             - cellSize - cell size can be set using this key in context parameter

                                             - extent - Sets the processing extent used by the function

                                             Setting context parameter will override the values set using arcgis.env
                                             variable for this particular function.(cellSize, extent)

                                             Example:

                                                {"exportAllTiles" : False, "startIndex": 0 }
    ------------------------------------     --------------------------------------------------------------------
    input_mask_polygons                      Optional :class:`~arcgis.features.FeatureLayer`. The feature layer that delineates the area where
                                             image chips will be created.
                                             Only image chips that fall completely within the polygons will be created.
    ------------------------------------     --------------------------------------------------------------------
    rotation_angle                           Optional float. The rotation angle that will be used to generate additional
                                             image chips.

                                             An image chip will be generated with a rotation angle of 0, which
                                             means no rotation. It will then be rotated at the specified angle to
                                             create an additional image chip. The same training samples will be
                                             captured at multiple angles in multiple image chips for data augmentation.
                                             The default rotation angle is 0.
    ------------------------------------     --------------------------------------------------------------------
    reference_system                         Optional string. Specifies the type of reference system to be used to interpret
                                             the input image. The reference system specified should match the reference system
                                             used to train the deep learning model.

                                             - MAP_SPACE : The input image is in a map-based coordinate system. This is the default.

                                             - IMAGE_SPACE : The input image is in image space, viewed from the direction of the sensor
                                               that captured the image, and rotated such that the tops of buildings and trees point upward in the image.

                                             - PIXEL_SPACE : The input image is in image space, with no rotation and no distortion.
    ------------------------------------     --------------------------------------------------------------------
    process_all_raster_items                 Optional bool. Specifies how all raster items in an image service will be processed.

                                             - False : all raster items in the image service will be mosaicked together and processed. This is the default.

                                             - True : all raster items in the image service will be processed as separate images.
    ------------------------------------     --------------------------------------------------------------------
    blacken_around_feature                   Optional bool. Specifies whether to blacken the pixels around each object or feature in each image tile.
                                             This parameter only applies when the metadata format is set to Labeled_Tiles and an input feature class or classified raster has been specified.

                                             - False : Pixels surrounding objects or features will not be blackened. This is the default.

                                             - True : Pixels surrounding objects or features will be blackened.

    ------------------------------------     --------------------------------------------------------------------
    fix_chip_size                            Optional bool. Specifies whether to crop the exported tiles such that they are all the same size.
                                             This parameter only applies when the metadata format is set to Labeled_Tiles and an input feature class or classified raster has been specified.

                                             - True : Exported tiles will be the same size and will center on the feature. This is the default.

                                             - False : Exported tiles will be cropped such that the bounding geometry surrounds only the feature in the tile.
    ------------------------------------     --------------------------------------------------------------------
    additional_input_raster                  Optional :class:`~arcgis.raster.ImageryLayer`/:class:`~arcgis.raster.Raster`/:class:`~arcgis.gis.Item`/String (URL).
                                             An additional input imagery source that will be used for image translation methods.

                                             This parameter is valid when the metadata_format parameter is set to Classified_Tiles, Export_Tiles, or CycleGAN.
    ------------------------------------     --------------------------------------------------------------------
    input_instance_data                      Optional. The training sample data collected that contains classes for instance segmentation.

                                             The input can also be a point feature without a class value field or an integer raster without any class information.

                                             This parameter is only valid when the metadata_format parameter is set to Panoptic_Segmentation.
    ------------------------------------     --------------------------------------------------------------------
    instance_class_value_field               Optional string. The field that contains the class values for instance segmentation.
                                             If no field is specified, the tool will use a value or class value field, if one is present.
                                             If the feature does not contain a class field, the tool will determine that all records belong to one class.

                                             This parameter is only valid when the metadata_format parameter is set to Panoptic_Segmentation.
    ------------------------------------     --------------------------------------------------------------------
    min_polygon_overlap_ratio                Optional float. The minimum overlap percentage for a feature to be included in the training data.
                                             If the percentage overlap is less than the value specified, the feature will be excluded from the
                                             training chip, and will not be added to the label file.

                                             The percent value is expressed as a decimal. For example, to specify an overlap of 20 percent,
                                             use a value of 0.2. The default value is 0, which means that all features will be included.

                                             This parameter improves the performance of the tool and also improves inferencing.
                                             The speed is improved since less training chips are created. Inferencing is improved
                                             since the model is trained to only detect large patches of objects and ignores small
                                             corners of features.

                                             This parameter is honoured only when the input_class_data parameter value is a feature service.
    ------------------------------------     --------------------------------------------------------------------
    gis                                      Optional :class:`~arcgis.gis.GIS` . The GIS on which this tool runs. If not specified, the active GIS is used.
    ------------------------------------     --------------------------------------------------------------------
    future                                   Keyword only parameter. Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    ====================================     ====================================================================

    :return:

        Output string containing the location of the exported training data

    """

    # task = "ExportTrainingDataforDeepLearning"

    gis = _arcgis.env.active_gis if gis is None else gis

    if gis._con._product == "AGOL":
        raise RuntimeError(
            "ArcGIS Online does not support export_training_data function."
        )

    return gis._tools.rasteranalysis.export_training_data_for_deep_learning(
        input_raster=input_raster,
        input_class_data=input_class_data,
        chip_format=chip_format,
        tile_size=tile_size,
        stride_size=stride_size,
        metadata_format=metadata_format,
        class_value_field=classvalue_field,
        buffer_radius=buffer_radius,
        output_location=output_location,
        input_mask_polygons=input_mask_polygons,
        rotation_angle=rotation_angle,
        reference_system=reference_system,
        process_all_raster_items=process_all_raster_items,
        blacken_around_feature=blacken_around_feature,
        fix_chip_size=fix_chip_size,
        additional_input_raster=additional_input_raster,
        input_instance_data=input_instance_data,
        instance_class_value_field=instance_class_value_field,
        min_polygon_overlap_ratio=min_polygon_overlap_ratio,
        context=context,
        future=future,
        **kwargs,
    )

    """
    url = gis.properties.helperServices.rasterAnalytics.url
    gptool = _arcgis.gis._GISResource(url, gis)

    params = {}

    if output_location:
        params["outputLocation"] = output_location
    else:
        raise RuntimeError("output_location cannot be None")

    if input_raster:
        params["inputRaster"] = _layer_input(input_raster)
    else:
        raise RuntimeError("input_raster cannot be None")

    if input_class_data:
        params["inputClassData"] = _layer_input(input_class_data)

    if chip_format is not None:
        chipFormatAllowedValues = ['TIFF', 'PNG', 'JPEG','MRF']
        if not chip_format in chipFormatAllowedValues:
            raise RuntimeError('chip_format can only be one of the following: '+ str(chipFormatAllowedValues))
        params["chipFormat"] = chip_format

    if tile_size:
        params["tileSize"] = tile_size

    if stride_size:
        params["strideSize"] = stride_size

    if metadata_format is not None:
        metadataFormatAllowedValues = ['KITTI_rectangles', 'PASCAL_VOC_rectangles', 'Classified_Tiles', 'RCNN_Masks', 'Labeled_Tiles']
        if not metadata_format in metadataFormatAllowedValues:
            raise RuntimeError('metadata_format can only be one of the following: '+ str(metadataFormatAllowedValues))

        params['metadataFormat'] = metadata_format

    if buffer_radius is not None:
        params["bufferRadius"]= buffer_radius

    if classvalue_field is not None:
        params["classValueField"]= classvalue_field

    if input_mask_polygons:
        params["inputMaskPolygons"] = _layer_input(input_mask_polygons)

    if rotation_angle:
        params["rotationAngle"] = rotation_angle

    _set_context(params, context)

    task_url, job_info, job_id = _analysis_job(gptool, task, params)

    job_info = _analysis_job_status(gptool, task_url, job_info)
    job_values = _analysis_job_results(gptool, task_url, job_info, job_id)
    item_properties = {
        "properties": {
            "jobUrl": task_url + '/jobs/' + job_info['jobId'],
            "jobType": "GPServer",
            "jobId": job_info['jobId'],
            "jobStatus": "completed"
        }
    }
    return job_values["outLocation"]["uri"]

    """


def list_models(*, gis=None, future=False, **kwargs):
    """
    Function is used to list all the installed deep learning models.

    .. note::
            This function is supported with ArcGIS Enterprise (Image Server)

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    gis                    Optional :class:`~arcgis.gis.GIS` . The GIS on which this tool runs. If not specified, the active GIS is used.
    ------------------     --------------------------------------------------------------------
    future                 Keyword only parameter. Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    ==================     ====================================================================

    :return:
        list of deep learning models installed

    """

    # task = "ListDeepLearningModels"

    gis = _arcgis.env.active_gis if gis is None else gis
    if gis._con._product == "AGOL":
        raise RuntimeError("ArcGIS Online does not support list_models function.")
    return gis._tools.rasteranalysis.list_deep_learning_models(future=future, **kwargs)
    """
    url = gis.properties.helperServices.rasterAnalytics.url
    gptool = _arcgis.gis._GISResource(url, gis)
    params = {}
    task_url, job_info, job_id = _analysis_job(gptool, task, params)

    job_info = _analysis_job_status(gptool, task_url, job_info)
    job_values = _analysis_job_results(gptool, task_url, job_info, job_id)
    item_properties = {
        "properties": {
            "jobUrl": task_url + '/jobs/' + job_info['jobId'],
            "jobType": "GPServer",
            "jobId": job_info['jobId'],
            "jobStatus": "completed"
        }
    }

    output_model_list = []
    if isinstance(job_values["deepLearningModels"], list) and job_values["deepLearningModels"] is not None:
        for element in job_values["deepLearningModels"]:
            if isinstance(element,dict):
                if "id" in element.keys():
                    item = gis.content.get(element["id"])
                    output_model_list.append(Model(item))
    return output_model_list
    """


def classify_objects(
    input_raster,
    model,
    model_arguments=None,
    input_features=None,
    class_label_field=None,
    process_all_raster_items=False,
    output_name=None,
    context=None,
    *,
    gis=None,
    future=False,
    **kwargs,
):
    """
    Function can be used to output feature service with assigned class label for each feature based on
    information from overlapped imagery data using the designated deep learning model.

    .. note::
            This function is supported with ArcGIS Enterprise (Image Server) and ArcGIS Image for ArcGIS Online.

    ====================================     ====================================================================
    **Parameter**                             **Description**
    ------------------------------------     --------------------------------------------------------------------
    input_raster                             Required. raster layer that contains objects that needs to be classified.
    ------------------------------------     --------------------------------------------------------------------
    model                                    Required :class:`~arcgis.learn.Model` object.
    ------------------------------------     --------------------------------------------------------------------
    model_arguments                          Optional dictionary. Name-value pairs of arguments and their values that can be customized by the clients.

                                             eg: {"name1":"value1", "name2": "value2"}
    ------------------------------------     --------------------------------------------------------------------
    input_features                           Optional :class:`~arcgis.features.FeatureLayer`.
                                             The point, line, or polygon input feature layer that identifies the location of each object to be
                                             classified and labelled. Each row in the input feature layer represents a single object.

                                             If no input feature layer is specified, the function assumes that each input image contains a single object
                                             to be classified. If the input image or images use a spatial reference, the output from the function is a
                                             feature layer, where the extent of each image is used as the bounding geometry for each labelled
                                             feature layer. If the input image or images are not spatially referenced, the output from the function
                                             is a table containing the image ID values and the class labels for each image.
    ------------------------------------     --------------------------------------------------------------------
    class_label_field                        Optional str. The name of the field that will contain the classification label in the output feature layer.

                                             If no field name is specified, a new field called ClassLabel will be generated in the output feature layer.

                                             Example:
                                                "ClassLabel"
    ------------------------------------     --------------------------------------------------------------------
    process_all_raster_items                 Optional bool.

                                             - If set to False, all raster items in the image service will be mosaicked together and processed. This is the default.

                                             - If set to True, all raster items in the image service will be processed as separate images.
    ------------------------------------     --------------------------------------------------------------------
    output_name                              Optional. If not provided, a :class:`~arcgis.features.FeatureLayer` is created by the method and used as the output .
                                             You can pass in an existing Feature Service Item from your GIS to use that instead.
                                             Alternatively, you can pass in the name of the output Feature Service that should be created by this method
                                             to be used as the output for the tool.
                                             A RuntimeError is raised if a service by that name already exists
    ------------------------------------     --------------------------------------------------------------------
    context                                  Optional dictionary. Context contains additional settings that affect task execution.
                                             Dictionary can contain value for following keys:

                                             - cellSize - Set the output raster cell size, or resolution

                                             - extent - Sets the processing extent used by the function

                                             - parallelProcessingFactor - Sets the parallel processing factor. Default is "80%"

                                             - processorType - Sets the processor type. "CPU" or "GPU"

                                             Eg: {"processorType" : "CPU"}

                                             Setting context parameter will override the values set using arcgis.env
                                             variable for this particular function.
    ------------------------------------     --------------------------------------------------------------------
    gis                                      Optional :class:`~arcgis.gis.GIS` . The GIS on which this tool runs. If not specified, the active GIS is used.
    ====================================     ====================================================================

    :return:
        The output feature layer item containing the classified objects

    """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.rasteranalysis.classify_objects_using_deep_learning(
        input_raster=input_raster,
        input_features=input_features,
        output_feature_class=output_name,
        model=model,
        model_arguments=model_arguments,
        class_label_field=class_label_field,
        process_all_raster_items=process_all_raster_items,
        context=context,
        future=future,
        **kwargs,
    )


def compute_accuracy_for_object_detection(
    detected_features,
    ground_truth_features,
    detected_class_value_field=None,
    ground_truth_class_value_field=None,
    min_iou=None,
    mask_features=None,
    out_accuracy_table_name=None,
    out_accuracy_report_name=None,
    context=None,
    *,
    gis=None,
    future=False,
    **kwargs,
):
    """
    Function can be used to calculate the accuracy of a deep learning model by comparing the detected objects from
    the detect_objects function to ground truth data.
    Function available in ArcGIS Image Server 10.9 and higher (not available in ArcGIS Online).

    ====================================     ====================================================================
    **Parameter**                             **Description**
    ------------------------------------     --------------------------------------------------------------------
    detected_features                        Required. The input polygon feature layer containing the objects
                                             detected from the detect_objects function.
    ------------------------------------     --------------------------------------------------------------------
    ground_truth_features                    Required. The polygon feature layer containing ground truth data.
    ------------------------------------     --------------------------------------------------------------------
    detected_class_value_field               Optional dictionary. The field in the detected objects feature class
                                             that contains the class names or class values.

                                             If a field name is not specified, a Classvalue or Value field will
                                             be used. If these fields do not exist, all records will be
                                             identified as belonging to one class.

                                             The class values or class names must match those in the ground truth feature class exactly.

                                             Syntax: A string describing the detected class value field.

                                             Example: "class"
    ------------------------------------     --------------------------------------------------------------------
    ground_truth_class_value_field           The field in the ground truth feature class that contains the class
                                             names or class values.

                                             If a field name is not specified, a Classvalue or Value field will
                                             be used. If these fields do not exist, all records will be
                                             identified as belonging to one class.

                                             The class values or class names must match those in the detected objects feature class exactly.

                                             Example: "class"
    ------------------------------------     --------------------------------------------------------------------
    min_iou                                  The Intersection over Union (IoU) ratio to use as a threshold to
                                             evaluate the accuracy of the object-detection model. The numerator
                                             is the area of overlap between the predicted bounding box and
                                             the ground truth bounding box. The denominator is the area of
                                             union or the area encompassed by both bounding boxes.

                                             min_IoU value should be in the range 0 to 1. [0,1]
                                             Example:
                                                0.5
    ------------------------------------     --------------------------------------------------------------------
    mask_features                            Optional :class:`~arcgis.features.FeatureLayer`. A polygon feature service layer that delineates
                                             the area where accuracy will be computed. Only the image area that
                                             falls completely within the polygons will be assessed for accuracy.
    ------------------------------------     --------------------------------------------------------------------
    out_accuracy_table_name                  Optional. Name of the output accuracy table item to be created.
                                             If not provided, a random name is generated by the method and used as
                                             the output name.
    ------------------------------------     --------------------------------------------------------------------
    out_accuracy_report_name                 Optional. Accuracy report can either be added as an item to the portal.
                                             or can be written to a datastore.
                                             To add as an item, specify the name of the output report item (pdf item)
                                             to be created.
                                             Example:

                                                "accuracyReport"

                                             In order to write accuracy report to datastore, specify the datastore path as value to uri key.

                                             Example -
                                                "/fileShares/yourFileShareFolderName/accuracyReport"
    ------------------------------------     --------------------------------------------------------------------
    context                                  Optional dictionary. Context contains additional settings that affect task execution.
                                             Dictionary can contain value for following keys:

                                             - cellSize - Set the output raster cell size, or resolution

                                             - extent - Sets the processing extent used by the function

                                             - parallelProcessingFactor - Sets the parallel processing factor. Default is "80%"

                                             - processorType - Sets the processor type. "CPU" or "GPU"

                                             Eg: {"processorType" : "CPU"}

                                             Setting context parameter will override the values set using arcgis.env
                                             variable for this particular function.
    ------------------------------------     --------------------------------------------------------------------
    gis                                      Optional :class:`~arcgis.gis.GIS` . The GIS on which this tool runs. If not specified, the active GIS is used.
    ====================================     ====================================================================

    :return:
        The output accuracy table item or/and accuracy report item (or datastore path to accuracy report)

    .. code-block:: python

        # Usage Example: This example generates an accuracy table for a specified minimum IoU value.

        compute_accuracy_op = compute_accuracy_for_object_detection(detected_features=detected_features,
                                                                    ground_truth_features=ground_truth_features,
                                                                    detected_class_value_field="ClassValue",
                                                                    ground_truth_class_value_field="Class",
                                                                    min_iou=0.5,
                                                                    mask_features=None,
                                                                    out_accuracy_table_name="accuracy_table",
                                                                    out_accuracy_report_name="accuracy_report",
                                                                    gis=gis)

    """

    gis = _arcgis.env.active_gis if gis is None else gis

    if gis._con._product == "AGOL":
        raise RuntimeError(
            "compute_accuracyfor_object_detection() is not supported on ArcGIS Online"
        )

    return gis._tools.rasteranalysis.compute_accuracyfor_object_detection(
        detected_features=detected_features,
        ground_truth_features=ground_truth_features,
        detected_class_value_field=detected_class_value_field,
        ground_truth_class_value_field=ground_truth_class_value_field,
        min_iou=min_iou,
        mask_features=mask_features,
        out_accuracy_table_name=out_accuracy_table_name,
        out_accuracy_report_name=out_accuracy_report_name,
        context=context,
        future=future,
        **kwargs,
    )


def train_model(
    input_folder,
    model_type,
    model_arguments=None,
    batch_size=2,
    max_epochs=None,
    learning_rate=None,
    backbone_model=None,
    validation_percent=None,
    pretrained_model=None,
    stop_training=True,
    freeze_model=True,
    overwrite_model=False,
    output_name=None,
    context=None,
    *,
    gis=None,
    future=False,
    **kwargs,
):
    """
    Function can be used to train a deep learning model using the output from the
    export_training_data function.
    It generates the deep learning model package (*.dlpk) and adds it to your enterprise portal.
    train_model function performs the training using the Raster Analytics server.

    .. note::
            This function is supported with ArcGIS Enterprise (Image Server)

    ====================================     ====================================================================
    **Parameter**                             **Description**
    ------------------------------------     --------------------------------------------------------------------
    input_folder                             Required string or list. This is the input location for the training sample data.
                                             It can be the path of output location on the file share raster data store or a
                                             shared file system path.
                                             The training sample data folder needs to be the output of export_training_data function,
                                             containing "images" and "labels" folder,
                                             as well as the JSON model definition file written out together by the function.

                                             File share raster store and datastore path examples:
                                               -  /rasterStores/yourRasterStoreFolderName/trainingSampleData
                                               - /fileShares/yourFileShareFolderName/trainingSampleData

                                             Shared path example:
                                               - \\serverName\deepLearning\trainingSampleData

                                             The function also support multiple input folders. In this case,
                                             specify the list of input folders


                                             list of file share raster store and datastore path examples:
                                               -  ["/rasterStores/yourRasterStoreFolderName/trainingSampleDataA", "/rasterStores/yourRasterStoreFolderName/trainingSampleDataB"]
                                               - ["/fileShares/yourFileShareFolderName/trainingSampleDataA", "/fileShares/yourFileShareFolderName/trainingSampleDataB"]

                                             list of shared path example:
                                               - ["\\serverName\deepLearning\trainingSampleDataA", "\\serverName\deepLearning\trainingSampleDataB"]

                                             Multiple input folders are supported when all the following conditions are met:

                                             - The metadata format must be one of the following types: Classified_Tiles, Labeled_Tiles, Multi-labeled Tiles, PASCAL_VOC_rectangles, or RCNN_Masks.
                                             - All training data must have the same metadata format.
                                             - All training data must have the same number of bands.
                                             - All training data must have the same tile size.
    ------------------------------------     --------------------------------------------------------------------
    model_type                               Required string. The model type to use for training the deep learning model.
                                             Possible values:

                                             - SSD - The Single Shot Detector (SSD) is used for object detection.
                                             - UNET - U-Net is used for pixel classification.
                                             - FEATURE_CLASSIFIER - The Feature Classifier is used for object classification.
                                             - PSPNET - The Pyramid Scene Parsing Network (PSPNET) is used for pixel classification.
                                             - RETINANET - The RetinaNet is used for object detection.
                                             - MASKRCNN - The MarkRCNN is used for object detection
                                             - YOLOV3 - The YOLOv3 approach will be used to train the model. YOLOv3 is used for object detection.
                                             - DeepLabV3 - The DeepLabV3 approach will be used to train the model. DeepLab is used for pixel classification.
                                             - FASTERRCNN - The FasterRCNN approach will be used to train the model. FasterRCNN is used for object detection.
                                             - BDCN_EDGEDETECTOR -  The Bi-Directional Cascade Network (BDCN) architecture will be used to train the model.
                                               The BDCN Edge Detector is used for pixel classification. This approach is useful to improve edge detection for objects at different scales.
                                             - HED_EDGEDETECTOR -  The Holistically-Nested Edge Detection (HED) architecture will be used to train the model.
                                               The HED Edge Detector is used for pixel classification. This approach is useful to in edge and object boundary detection.
                                             - MULTITASK_ROADEXTRACTOR -  The Multi Task Road Extractor architecture will be used to train the model.
                                               The Multi Task Road Extractor is used for pixel classification. This approach is useful for road network extraction from satellite imagery.
                                             - CONNECTNET - The ConnectNet architecture will be used to train the model. ConnectNet is used for pixel classification.
                                               This approach is useful for road network extraction from satellite imagery.
                                             - PIX2PIX - The Pix2Pix approach will be used to train the model. Pix2Pix is used for image-to-image translation.
                                               This approach creates a model object that generates images of one type to another. The input training data for this
                                               model type uses the Export Tiles metadata format.
                                             - CYCLEGAN - The CycleGAN approach will be used to train the model. CycleGAN is used for image-to-image translation.
                                               This approach creates a model object that generates images of one type to another. This approach is unique in that
                                               the images to be trained do not need to overlap. The input training data for this model type uses the CycleGAN metadata format.
                                             - SUPERRESOLUTION - The Super-resolution approach will be used to train the model. Super-resolution is used for
                                               image-to-image translation. This approach creates a model object that increases the resolution and improves the
                                               quality of images. The input training data for this model type uses the Export Tiles metadata format.
                                             - CHANGEDETECTOR - The Change detector approach will be used to train the model. Change detector is used for
                                               pixel classification. This approach creates a model object that uses two spatial-temporal images to create
                                               a classified raster of the change. The input training data for this model type uses the Classified Tiles metadata format.
                                             - IMAGECAPTIONER - The Image captioner approach will be used to train the model. Image captioner is used for
                                               image-to-text translation. This approach creates a model that generates text captions for an image.
                                             - SIAMMASK - The Siam Mask approach will be used to train the model. Siam Mask is used for object detection in videos.
                                               The model is trained using frames of the video and detects the classes and bounding boxes of the objects in each frame.
                                               The input training data for this model type uses the MaskRCNN metadata format.
                                             - MMDETECTION - The MMDetection approach will be used to train the model. MMDetection is used for object detection.
                                               The supported metadata formats are PASCAL Visual Object Class rectangles and KITTI rectangles.
                                             - MMSEGMENTATION - The MMSegmentation approach will be used to train the model. MMDetection is used for pixel classification.
                                               The supported metadata format is Classified Tiles.
                                             - DEEPSORT - The Deep Sort approach will be used to train the model. Deep Sort is used for object detection in videos.
                                               The model is trained using frames of the video and detects the classes and bounding boxes of the objects in each frame.
                                               The input training data for this model type uses the Imagenet metadata format.
                                               Where Siam Mask is useful while tracking an object, Deep Sort is useful in training a model to track multiple objects.
                                             - PIX2PIXHD - The Pix2PixHD approach will be used to train the model. Pix2PixHD is used for image-to-image translation.
                                               This approach creates a model object that generates images of one type to another.
                                               The input training data for this model type uses the Export Tiles metadata format.
                                             - MAXDEEPLAB - The MAXDEEPLAB approach will be used to train the model. It is used for Panoptic Segmentation.
    ------------------------------------     --------------------------------------------------------------------
    model_arguments                          Optional dictionary. Name-value pairs of arguments and their values that can be customized by the clients.

                                             Example:
                                                {"name1":"value1", "name2": "value2"}
    ------------------------------------     --------------------------------------------------------------------
    batch_size                               Optional int.
                                             The number of training samples to be processed for training at one time.
                                             If the server has a powerful GPU, this number can be increased to 16, 36, 64, and so on.

                                             Example:
                                                4
    ------------------------------------     --------------------------------------------------------------------
    max_epochs                               Optional int. The maximum number of epochs that the model should be trained.
                                             One epoch means the whole training dataset will be passed forward and backward
                                             through the deep neural network once.

                                             Example:
                                                20
    ------------------------------------     --------------------------------------------------------------------
    learning_rate                            Optional float.
                                             The rate at which the weights are updated during the training.
                                             It is a small positive value in the range between 0.0 and 1.0.
                                             If learning rate is set to 0, it will extract the optimal learning rate
                                             from the learning curve during the training process.

                                             Example:
                                                0.0
    ------------------------------------     --------------------------------------------------------------------
    backbone_model                           Optional string.
                                             Specifies the preconfigured neural network to be used as an architecture for training the new model.
                                             Possible values: DENSENET121 , DENSENET161 , DENSENET169 , DENSENET201 , MOBILENET_V2 ,
                                             RESNET18 , RESNET34 , RESNET50 , RESNET101 , RESNET152 , VGG11 , VGG11_BN , VGG13 ,
                                             VGG13_BN , VGG16 , VGG16_BN , VGG19 , VGG19_BN , DARKNET53 , REID_V1 , REID_V2

                                             Example:
                                                RESNET34
    ------------------------------------     --------------------------------------------------------------------
    validation_percent                       Optional float.
                                             The percentage (in %) of training sample data that will be used for validating the model.

                                             Example:
                                                10
    ------------------------------------     --------------------------------------------------------------------
    pretrained_model                         Optional dlpk portal item.

                                             The pretrained model to be used for fine tuning the new model.
                                             It is a deep learning model package (dlpk) portal item.
    ------------------------------------     --------------------------------------------------------------------
    stop_training                            Optional bool.
                                             Specifies whether early stopping will be implemented.

                                             - True - The model training will stop when the model is no longer improving,
                                               regardless of the maximum epochs specified. This is the default.
                                             - False - The model training will continue until the maximum epochs is reached.
    ------------------------------------     --------------------------------------------------------------------
    freeze_model                             Optional bool.
                                             Specifies whether to freeze the backbone layers in the pretrained model,
                                             so that the weights and biases in the backbone layers remain unchanged.

                                             - True - The predefined weights and biases will not be altered in the backboneModel.
                                               This is the default.
                                             - False - The weights and biases of the backboneModel may be altered to better
                                               fit your training samples. This may take more time to process but
                                               usually could get better results.
    ------------------------------------     --------------------------------------------------------------------
    overwrite_model                          Optional bool.
                                             Overwrites an existing deep learning model package (.dlpk) portal item with the same name.

                                             If the output_name parameter uses the file share data store path, this overwriteModel parameter is not applied.

                                             - True - The portal .dlpk item will be overwritten.
                                             - False - The portal .dlpk item will not be overwritten. This is the default.
    ------------------------------------     --------------------------------------------------------------------
    output_name                              Optional. trained deep learning model package can either be added as an item
                                             to the portal or can be written to a datastore.

                                             To add as an item, specify the name of the output deep learning model package (item)
                                             to be created.

                                             Example -
                                                "trainedModel"

                                             In order to write the dlpk to fileshare datastore, specify the datastore path.

                                             Example -
                                                "/fileShares/filesharename/folder"
    ------------------------------------     --------------------------------------------------------------------
    context                                  Optional dictionary. Context contains additional settings that affect task execution.
                                             Dictionary can contain value for following keys:

                                             - cellSize - Set the output raster cell size, or resolution
                                             - extent - Sets the processing extent used by the function
                                             - parallelProcessingFactor - Sets the parallel processing factor. Default is "80%"
                                             - processorType - Sets the processor type. "CPU" or "GPU"
                                             Example -
                                                {"processorType" : "CPU"}

                                             Setting context parameter will override the values set using arcgis.env
                                             variable for this particular function.
    ------------------------------------     --------------------------------------------------------------------
    gis                                      Optional :class:`~arcgis.gis.GIS` . The GIS on which this tool runs. If not specified, the active GIS is used.
    ====================================     ====================================================================

    :return:
        Returns the dlpk portal item that has properties for title, type, filename, file, id and folderId.
    """

    gis = _arcgis.env.active_gis if gis is None else gis

    if gis._con._product == "AGOL":
        raise RuntimeError("ArcGIS Online does not support train_model function.")

    return gis._tools.rasteranalysis.train_deep_learning_model(
        in_folder=input_folder,
        output_name=output_name,
        model_type=model_type,
        arguments=model_arguments,
        batch_size=batch_size,
        max_epochs=max_epochs,
        learning_rate=learning_rate,
        backbone_model=backbone_model,
        validation_percent=validation_percent,
        pretrained_model=pretrained_model,
        stop_training=stop_training,
        freeze_model=freeze_model,
        overwrite_model=overwrite_model,
        context=context,
        future=future,
        **kwargs,
    )


def detect_change_using_deep_learning(
    from_raster,
    to_raster,
    model,
    output_classified_raster=None,
    model_arguments=None,
    context=None,
    *,
    gis=None,
    future=False,
    **kwargs,
):
    """
    Runs a trained deep learning model to detect change between two rasters.
    Function available in ArcGIS Image Server 11.1 and higher.

    ====================================     ====================================================================
    **Argument**                             **Description**
    ------------------------------------     --------------------------------------------------------------------
    from_raster                              Required ImageryLayer object. The previous raster to use for change detection.
    ------------------------------------     --------------------------------------------------------------------
    to_raster                                Required ImageryLayer object. The recent raster to use for change detection.
    ------------------------------------     --------------------------------------------------------------------
    model                                    Required. The deep learning model to be used for the change detection.
                                             It can be passed as a dlpk portal item, datastore path to the Esri Model Definition (EMD)
                                             file or the EMD JSON string.
    ------------------------------------     --------------------------------------------------------------------
    output_classified_raster                 Optional String. If not provided, an Image Service is created by the method and used as the output raster.
                                             You can pass in an existing Image Service Item from your GIS to use that instead.

                                             Alternatively, you can pass in the name of the output Image Service that should be created by this method to be
                                             used as the output for the tool.

                                             A RuntimeError is raised if a service by that name already exists.
    ------------------------------------     --------------------------------------------------------------------
    model_arguments                          Optional dictionary. Name-value pairs of arguments and their values that can be customized by the clients.

                                             eg: {"name1":"value1", "name2": "value2"}
    ------------------------------------     --------------------------------------------------------------------
    context                                  Context contains additional settings that affect task execution.

                                                context parameter overwrites values set through arcgis.env parameter

                                                This function has the following settings:

                                                - Cell size (cellSize) - Set the output raster cell size, or resolution

                                                - Output Spatial Reference (outSR): The output raster will be
                                                projected into the output spatial reference.

                                                Example:
                                                    {"outSR": {spatial reference}}

                                                - Extent (extent): A bounding box that defines the analysis area.

                                                Example:
                                                    {"extent": {"xmin": -122.68,
                                                    "ymin": 45.53,
                                                    "xmax": -122.45,
                                                    "ymax": 45.6,
                                                    "spatialReference": {"wkid": 4326}}}

                                                - Parallel Processing Factor (parallelProcessingFactor): controls
                                                Raster Processing (CPU) service instances.

                                                Example:
                                                    Syntax example with a specified number of processing instances:

                                                    {"parallelProcessingFactor": "2"}

                                                    Syntax example with a specified percentage of total
                                                    processing instances:

                                                    {"parallelProcessingFactor": "60%"}
    ------------------------------------     --------------------------------------------------------------------
    gis                                      Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    ------------------------------------     --------------------------------------------------------------------
    future                                   Keyword only parameter. Optional Boolean. If True, the result will be a GPJob object and
                                             results will be returned asynchronously.
    ------------------------------------     --------------------------------------------------------------------
    folder                                   Keyword only parameter. Optional str or dict. Creates a folder in the portal, if it does
                                             not exist, with the given folder name and persists the output in this folder.
                                             The dictionary returned by the gis.content.create_folder() can also be passed in as input.

                                             Example:
                                                {'username': 'user1', 'id': '6a3b77c187514ef7873ba73338cf1af8', 'title': 'trial'}
    ====================================     ====================================================================

    :return: The output imagery layer item

    .. code-block:: python

        # Usage Example 1:

        from_raster = gis.content.search("from_raster", item_type="Imagery Layer")[0].layers[0]
        to_raster = gis.content.search("to_raster", item_type="Imagery Layer")[0].layers[0]
        change_detection_model = gis.content.search("my_detection_model")[0]

        detect_change_op = detect_change_using_deep_learning(from_raster=from_raster,
                                                             to_raster=to_raster,
                                                             model=change_detection_model,
                                                             gis=gis)

    """

    gis = _arcgis.env.active_gis if gis is None else gis

    return gis._tools.rasteranalysis.detect_change_using_deep_learning(
        from_raster=from_raster,
        to_raster=to_raster,
        model=model,
        output_classified_raster=output_classified_raster,
        model_arguments=model_arguments,
        context=context,
        future=future,
        **kwargs,
    )


class Model:
    def __init__(self, model=None):
        self._model_package = False
        if isinstance(model, _arcgis.gis.Item):
            self._model = _json.dumps({"itemId": model.itemid})
            self._model_package = True
            self.item = model

    def _repr_html_(self):
        if self._model_package:
            if hasattr(self, "item"):
                self.item._repr_html_()
        else:
            self.__repr__()

    def __repr__(self):
        if self._model_package:
            model = _json.loads(self._model)
            if "url" in model.keys():
                return "<Model:%s>" % self._model
            return "<Model Title:%s owner:%s>" % (self.item.title, self.item.owner)

        else:
            try:
                return "<Model:%s>" % self._model
            except:
                return "<empty Model>"

    def from_json(self, model):
        """
        Function is used to initialize Model object from model definition JSON

        .. code-block:: python

            # Usage example

            >>> model = Model()

            >>> model.from_json({"Framework" :"TensorFlow",
                                 "ModelConfiguration":"DeepLab",
                                 "InferenceFunction":"``[functions]System\\DeepLearning\\ImageClassifier.py``",
                                 "ModelFile":"``\\\\folder_path_of_pb_file\\frozen_inference_graph.pb``",
                                 "ExtractBands":[0,1,2],
                                 "ImageWidth":513,
                                 "ImageHeight":513,
                                 "Classes": [ { "Value":0, "Name":"Evergreen Forest", "Color":[0, 51, 0] },
                                              { "Value":1, "Name":"Grassland/Herbaceous", "Color":[241, 185, 137] },
                                              { "Value":2, "Name":"Bare Land", "Color":[236, 236, 0] },
                                              { "Value":3, "Name":"Open Water", "Color":[0, 0, 117] },
                                              { "Value":4, "Name":"Scrub/Shrub", "Color":[102, 102, 0] },
                                              { "Value":5, "Name":"Impervious Surface", "Color":[236, 236, 236] } ] })

        """
        if isinstance(model, dict):
            self._model = model
            self._model_package = False

    def from_model_path(self, model):
        """
        Function is used to initialize Model object from url of model package or path of model definition file

        .. code-block:: python

            # Usage Example #1:

            >>> model = Model()
            >>> model.from_model_path("https://xxxportal.esri.com/sharing/rest/content/items/<itemId>")

            # Usage Example #2:

            >>> model = Model()
            >>> model.from_model_path("\\\\sharedstorage\\sharefolder\\findtrees.emd")
        """
        if "http:" in model or "https:" in model:
            self._model = _json.dumps({"url": model})
            self._model_package = True
        else:
            self._model = _json.dumps({"uri": model})
            self._model_package = False

    def install(self, *, gis=None, future=False, **kwargs):
        """
        Function is used to install the uploaded model package (*.dlpk). Optionally after inferencing
        the necessary information using the model, the model can be uninstalled by uninstall_model()


        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        gis                    Optional :class:`~arcgis.gis.GIS` . The GIS on which this tool runs. If not specified, the active GIS is used.
        ------------------     --------------------------------------------------------------------
        future                 Keyword only parameter. Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
        ==================     ====================================================================

        :return:
            Path where model is installed

        """
        if self._model_package is False:
            raise RuntimeError(
                "model object should be created from a portal item or a portal url"
            )

        if self._model is None:
            raise RuntimeError(
                "For install/uninstall model object should be created from a portal item or portal url"
            )

        # task = "InstallDeepLearningModel"

        gis = _arcgis.env.active_gis if gis is None else gis

        if gis._con._product == "AGOL":
            raise RuntimeError(
                "ArcGIS Online does not support install method on a Model Object. The Model object can be directly used with deep learning functions without installation."
            )
        return gis._tools.rasteranalysis.install_deep_learning_model(
            model_package=self._model, future=future, **kwargs
        )

        """
        url = gis.properties.helperServices.rasterAnalytics.url
        gptool = _arcgis.gis._GISResource(url, gis)

        params = {}

        if self._model is None:
            raise RuntimeError("For install/uninstall model object should be created from a portal item or portal url")
        else:
            params["modelPackage"] = self._model

        task_url, job_info, job_id = _analysis_job(gptool, task, params)

        job_info = _analysis_job_status(gptool, task_url, job_info)
        job_values = _analysis_job_results(gptool, task_url, job_info, job_id)
        item_properties = {
            "properties": {
                "jobUrl": task_url + '/jobs/' + job_info['jobId'],
                "jobType": "GPServer",
                "jobId": job_info['jobId'],
                "jobStatus": "completed"
            }
        }

        return job_values["installSucceed"]
        """

    def query_info(self, *, gis=None, future=False, **kwargs):
        """
        Function is used to extract the deep learning model specific settings from the model package item or model definition file.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        gis                    Optional :class:`~arcgis.gis.GIS` . The GIS on which this tool runs. If not specified, the active GIS is used.
        ------------------     --------------------------------------------------------------------
        future                 Keyword only parameter. Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
        ==================     ====================================================================

        :return:
           The key model information in dictionary format that describes what the settings are essential for this type of deep learning model.
        """

        # task = "QueryDeepLearningModelInfo"

        gis = _arcgis.env.active_gis if gis is None else gis
        if self._model is None:
            raise RuntimeError("model cannot be None")

        return gis._tools.rasteranalysis.query_deep_learning_model_info(
            model=self._model, future=future, **kwargs
        )

        """
        url = gis.properties.helperServices.rasterAnalytics.url
        gptool = _arcgis.gis._GISResource(url, gis)

        params = {}

        if self._model is None:
            raise RuntimeError('model cannot be None')
        else:
            params["model"] = self._model

        task_url, job_info, job_id = _analysis_job(gptool, task, params)

        job_info = _analysis_job_status(gptool, task_url, job_info)
        job_values = _analysis_job_results(gptool, task_url, job_info, job_id)
        item_properties = {
            "properties": {
                "jobUrl": task_url + '/jobs/' + job_info['jobId'],
                "jobType": "GPServer",
                "jobId": job_info['jobId'],
                "jobStatus": "completed"
            }
        }
        output = job_values["outModelInfo"]
        print(output)
        try:
            dict_output =  _json.loads(output["modelInfo"])
            return dict_output
        except:
            return output
        """

    def uninstall(self, *, gis=None, future=False, **kwargs):
        """
        Function is used to uninstall the uploaded model package that was installed using the install_model()
        This function will delete the named deep learning model from the server but not the portal item.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        gis                    Optional :class:`~arcgis.gis.GIS` . The GIS on which this tool runs. If not specified, the active GIS is used.
        ------------------     --------------------------------------------------------------------
        future                 Keyword only parameter. Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
        ==================     ====================================================================

        :return:
            itemId of the uninstalled model package item

        """
        if self._model_package is False:
            raise RuntimeError(
                "For install/uninstall model object should be created from a portal item or a portal url"
            )

        # task = "UninstallDeepLearningModel"

        gis = _arcgis.env.active_gis if gis is None else gis
        if gis._con._product == "AGOL":
            raise RuntimeError(
                "ArcGIS Online does not support uninstall method on a Model Object."
            )

        if self._model is None:
            raise RuntimeError("model_package cannot be None")

        return gis._tools.rasteranalysis.uninstall_deep_learning_model(
            model_item_id=self._model, future=future, **kwargs
        )

        """
        url = gis.properties.helperServices.rasterAnalytics.url
        gptool = _arcgis.gis._GISResource(url, gis)

        params = {}

        if self._model is None:
            raise RuntimeError('model_package cannot be None')
        else:
            params["modelItemId"] = self._model

        task_url, job_info, job_id = _analysis_job(gptool, task, params)

        job_info = _analysis_job_status(gptool, task_url, job_info)
        job_values = _analysis_job_results(gptool, task_url, job_info, job_id)
        item_properties = {
            "properties": {
                "jobUrl": task_url + '/jobs/' + job_info['jobId'],
                "jobType": "GPServer",
                "jobId": job_info['jobId'],
                "jobStatus": "completed"
            }
        }

        return job_values["uninstallSucceed"]
        """


def export_point_dataset(
    data_path,
    output_path,
    block_size=50.0,
    max_points=8192,
    extra_features=[],
    **kwargs,
):
    """
    Note:
    This function has been deprecated starting from `ArcGIS API for
    Python` version 1.9.0. Export data using `Prepare Point Cloud Training Data` tool available
    in 3D Analyst Extension from ArcGIS Pro 2.8 onwards.

    """

    from ._utils.pointcloud_data import prepare_las_data

    prepare_las_data(
        data_path, block_size, max_points, output_path, extra_features, **kwargs
    )



from arcgis import GIS

from .bigdata_analytics_manager import BigDataAnalyticsManager
from .feeds_manager import FeedsManager
from .realtime_analytics_manager import RealTimeAnalyticsManager
from ._util import _Util


class Velocity:
    """
    Provides access to real-time analytics, big data analytics, and feeds in ArcGIS Velocity.

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    URL of the ArcGIS Velocity organization.
    ------------------     --------------------------------------------------------------------
    gis                    an authenticated :class:`arcigs.gis.GIS` object.
    ==================     ====================================================================

    .. code-block:: python

        # Connect to a Velocity instance:

        gis = GIS(url="url",username="username",password="password",)

        velocity = gis.velocity
        velocity

    .. code-block:: python

        # UsageExample:

        from arcgis.gis import GIS
        gis = GIS(url="url",username="username",password="password",)

        velocity = gis.velocity
        velocity

    """

    _gis = None
    _url = None
    _subinfo = None
    _velocity = None
    # manager instances
    _feeds = None
    _realtime_analytics = None
    _bigdata_analytics = None
    _velocity = None
    _util = None

    def __init__(self, url: str, gis: GIS):
        self._gis = gis
        self._url = url
        # Optional may set gis and _url to Velocity if we need to access these
        # in other classes internally such as: Velocity._gis, Velocity._url
        Velocity._gis = gis
        Velocity._url = url
        # set reference for Util Velocity
        Velocity._util = _Util(gis, url)

    @property
    def feeds(self) -> FeedsManager:
        """
        Provides access to the resource manager for managing configured feeds in ArcGIS Velocity.

        :return: :class:`~arcgis.realtime.velocity.FeedsManager`

        .. code-block:: python

            # Get instance of feeds from `velocity`:

            feeds = velocity.feeds
            feeds

        """
        if self._feeds is None:
            self._feeds = FeedsManager(url=self._url, gis=self._gis)
        return self._feeds

    @property
    def realtime_analytics(self) -> RealTimeAnalyticsManager:
        """
         Provides access to the resource manager for managing configured real-time analytic tasks in ArcGIS Velocity.

        :return: :class:`~arcgis.realtime.velocity.RealTimeAnalyticsManager`

        .. code-block:: python

            # Get instance of realtime_analytics from `velocity`:

            realtime_analytics = velocity.realtime_analytics
            realtime_analytics
        """
        if self._realtime_analytics is None:
            self._realtime_analytics = RealTimeAnalyticsManager(
                url=self._url, gis=self._gis
            )
        return self._realtime_analytics

    @property
    def bigdata_analytics(self) -> BigDataAnalyticsManager:
        """
        Provides access to the resource manager for managing configured big data analytic tasks in ArcGIS Velocity.

        :return: :class:`~arcgis.realtime.velocity.BigDataAnalyticsManager`

        .. code-block:: python

            # Get instance of bigdata_analytics from `velocity`:

            bigdata_analytics = velocity.bigdata_analytics
            bigdata_analytics

        """
        if self._bigdata_analytics is None:
            self._bigdata_analytics = BigDataAnalyticsManager(
                url=self._url, gis=self._gis
            )
        return self._bigdata_analytics



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# cascade_rcnn_r2_101_fpn_20e_coco.py, box AP=45.7

_base_ = "./_base_/models/cascade_rcnn_r50_fpn.py"

model = dict(
    pretrained="open-mmlab://res2net101_v1d_26w_4s",
    backbone=dict(type="Res2Net", depth=101, scales=4, base_width=26),
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/res2net/cascade_rcnn_r2_101_fpn_20e_coco/cascade_rcnn_r2_101_fpn_20e_coco-f4b7b7db.pth"



"""
arcgis.realtime.StreamLayer provides types and functions for receiving real-time data feeds and sensor data streamed from
the GIS to perform continuous processing and analysis on the streaming data. It includes support for stream layers that allow Python scripts
to subscribe to the streamed feature data or to broadcast updates and alerts.

"""
from arcgis.gis import *
from arcgis.features import *

from urllib.parse import urlencode


class StreamLayer(Layer):
    """
    Allows Python scripts to subscribe to the feature data streamed from the GIS, using ArcGIS
    GeoEvent Server or ArcGIS Velocity, or to broadcast updates and alerts. This class can be used to perform continuous processing and
    analysis on streaming data as it is received.
    """

    # autobahn, twisted, pyOpenssl, service_identity
    def __init__(self, url, gis=None):
        super(StreamLayer, self).__init__(url, gis)
        self._streamtoken = self.properties.streamUrls[0].token
        self._streamurl = self.properties.streamUrls[0].urls[0]
        self._out_sr = self.properties.spatialReference.wkid
        self.filter = {}
        self._on_features = None
        self._on_disconnect = None
        self._on_error = None
        try:
            from arcgis.gis.server._service._adminfactory import AdminServiceGen

            self.service = AdminServiceGen(service=self, gis=gis)
        except:
            pass

    @property
    def out_sr(self):
        """Get/Set the spatial reference of the streamed features."""
        return self._out_sr

    @out_sr.setter
    def out_sr(self, value):
        self._out_sr = value

    @property
    def filter(self):
        """
        Get/Set property used for filtering the streamed features so they meet spatial and SQL like criteria,
        and return the specified fields.
        """
        return self._filter

    @filter.setter
    def filter(self, value):
        self._filter = value

    def subscribe(self, on_features, on_open=None, on_disconnect=None, on_error=None):
        """
        Allows Python scripts to subscribe to the feature data streamed from the GIS using ArcGIS
        GeoEvent Server or ArcGIS Velocity. Subscribing to the streamed data can be used to perform continuous processing and
        analysis of real-time data as it is received.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        on_features         callback function that is called every time features are streamed
                            to the client.
        ---------------     --------------------------------------------------------------------
        on_open             callback function called when the connection to the streaming server
                            is created.
        ---------------     --------------------------------------------------------------------
        on_disconnect       callback function called when the connection to the streaming server
                            is closed.
        ---------------     --------------------------------------------------------------------
        on_error            callback function called if the connection recieves an error.
        ===============     ====================================================================

        """
        try:
            import sys
            import ssl
            from twisted.internet import reactor
            from twisted.python import log

            from autobahn.twisted.websocket import (
                WebSocketClientFactory,
                WebSocketClientProtocol,
                connectWS,
            )
        except:
            raise ImportError(
                "Install autobahn, twisted, pyOpenssl, service_identity packages to subscribe"
            )

        url = self._streamurl

        params = {"token": self._streamtoken}
        params.update(self.filter)

        if self.out_sr != self.properties.spatialReference.wkid:
            params["outSR"] = self.out_sr

        url = "{url}/subscribe?{params}".format(url=url, params=urlencode(params))

        class StreamServiceClientProtocol(WebSocketClientProtocol):
            def onOpen(self):
                if on_open is not None:
                    on_open()

            def onMessage(self, payload, isBinary):
                if isBinary:
                    print("Binary message received: {0} bytes".format(len(payload)))
                else:
                    msg = format(payload.decode("utf8"))
                    on_features(msg)

        factory = WebSocketClientFactory(url, headers={"token": self._streamtoken})

        factory.protocol = StreamServiceClientProtocol
        connectWS(factory)

        reactor.run()
